import platform
from typing import Union, Optional, Iterable, Any
from types import MethodType
from dataclasses import dataclass
from hrenpack.decorators import args_kwargs
from hrenpack.listwork import split_list, intlist, floatlist, merging_dictionaries, if_dict_key, dict_keyf
from hrenpack.numwork import dec_to_hex, hex_to_dec


class stl(list):
    def __init__(self, source: Union[str, int, float, bool], *args, **kwargs) -> None:
        super().__init__()
        self.vstring = str(source)
        self.vlist = self.vstring.split(', ')
        self.save()
        self.__float__ = lambda: floatlist(self.vlist)
        self.__int__ = lambda: intlist(self.vlist)
        self.args, self.kwargs = args, kwargs
        self.not_empty = self.__bool__

    def __str__(self) -> str:
        return self.vstring

    def __list__(self) -> list:
        return self.vlist

    def __tuple__(self) -> tuple:
        return self.vtuple

    def save(self) -> None:
        self.vtuple = tuple(self.vlist)
        self.vstring = split_list(self.vlist)

    def split(self, isTuple: bool) -> Union[tuple, list]:
        return self.__tuple__() if isTuple else self.__list__()

    def append(self, value) -> None:
        self.vstring = f'{self.vstring}, {value}'
        self.vlist.append(value)
        self.save()

    def pop(self, index: int = -1) -> None:
        self.vlist.pop(index)
        self.save()

    def reverse(self) -> None:
        self.vlist.reverse()
        self.save()

    def remove(self, value) -> None:
        self.vlist.remove(value)
        self.save()

    def count(self, value) -> int:
        return self.vlist.count(value)

    def index(self, value, **kwargs) -> int:
        return self.vlist.index(value)

    def __len__(self) -> int:
        return len(self.vlist)

    def __bool__(self) -> bool:
        return bool(self.vlist)

    def __copy__(self):
        return stl(self.vstring)

    def __hash__(self):
        return hash(self.vstring)

    def __eq__(self, other) -> bool:
        return self.vstring == other.vstring

    def __ne__(self, other) -> bool:
        return self.vstring != other.vstring

    def clear(self):
        self.vlist.clear()
        self.save()

    def copy(self):
        return stl(self.vstring)

    def sort(self, *, key=None, reverse=False):
        self.vlist.sort(key=key, reverse=reverse)
        self.save()

    def __setitem__(self, key, value):
        self.vlist[key] = value
        self.save()

    def __getitem__(self, key):
        return self.vlist[key]

    def __delitem__(self, key):
        del self.vlist[key]
        self.save()

    def insert(self, __index, __object):
        self.vlist.insert(__index, __object)
        self.save()


class DictionaryWithExtendedFunctionality(dict):
    def __init__(self, **kwargs):
        super().__init__()
        self.__dict__ = kwargs

    def merge(self, *dicts):
        self.__dict__ = merging_dictionaries(self.__dict__, *dicts)


class MatrixCore:
    def __init__(self, path_to_file: str = '</return/>'):
        self.path_to_file = path_to_file
        self.is_return = self.path_to_file == '</return/>'

    class RectMatrix:
        def __init__(self, width: int, height: int, default_value=0):
            self.matrix = list()
            self.width, self.height = width, height
            for y in range(height):
                yl = list()
                for x in range(width):
                    yl.append(default_value)
                self.matrix.append(yl)

        def __setitem__(self, x: int, y: int, value):
            self.matrix[y][x] = value

        def __getitem__(self, x: int, y: int):
            return self.matrix[y][x]

        def __str__(self, separator: str = ' ') -> str:
            def step1(argument):
                def step2(arg):
                    def step3(a):
                        pass

                    vs = arg.split()
                    return split_list(vs, separator)

                al = argument.split('\n')
                al.pop(0)

                return split_list(al, '\n')

            output = ''
            for yl in self.matrix:
                po = ''
                for xel in yl:
                    po = po + separator + str(xel)
                output = output + '\n' + po
            return step1(output)


class DataClass:
    __default_classname__: str = 'DataClass'

    def __init__(self, **kwargs):
        self.__classname__ = kwargs.get('__classname__', self.__default_classname__)
        if '__classname__' in kwargs.keys():
            del kwargs['__classname__']
        self.__kwargs__: dict = kwargs
        self.__methods__ = ('__dict__', '__setitem__', '__getitem__', '__delitem__', '__len__', '__bool__', '__copy__',
                            '__update__', '__str__', '__clear__', '__methods__')
        self.__update__(**kwargs)

    def __dict__(self):
        return self.__kwargs__

    def __setitem__(self, key, value):
        self.__kwargs__[key] = value

    def __getitem__(self, key):
        return self.__kwargs__[key]

    def __delitem__(self, key):
        del self.__kwargs__[key]

    def __len__(self):
        return len(self.__kwargs__)

    def __bool__(self):
        return bool(self.__kwargs__)

    def __copy__(self):
        return DataClass(**self.kwargs)

    def __update__(self, **kwargs):
        for key, value in kwargs.items():
            if key not in self.__methods__:
                self.__kwargs__[key] = value
                setattr(self, key, value)
            else:
                raise ValueError(f"Имя {key} зарезервировано")

    # def __str__(self):
    #     classname = self.__classname__
    #     if classname == '</empty/>':
    #         string = ''
    #     else:
    #         string = classname + '('
    #     for key, value in self.__kwargs__.items():
    #         string += f'{key}={value}, '
    #     string = string[:-2]
    #     if classname != '</empty/>':
    #         string += ')'
    #     return string

    def __str__(self):
        string = self.__classname__ + '('
        for key, value in self.__kwargs__.items():
            string += f'{key}={value if type(value) is not str else f"'{value}'"}, '
        string = string[:-2]
        string += ')'
        return string

    def __clear__(self):
        self.__kwargs__.clear()

    def __ad_to_dt__(self):
        dicts = dict()
        for key, value in self.__kwargs__.items():
            if type(value) is dict:
                dicts[key] = PreEmptyDataClass(**value)
        self.__update__(**dicts)

    def __iter__(self):
        return iter(self.__kwargs__.items())


class PreEmptyDataClass(DataClass):
    __default_classname__ = ''

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if self.__classname__ != self.__default_classname__:
            self.__update__(__classname__=self.__classname__)


class EmptyDataClass(DataClass):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        try:
            del self.__classname__, self.__default_classname__
        except AttributeError:
            pass

    def __str__(self):
        string = ''
        for key, value in self.__kwargs__.items():
            string += f'{key}={value}, '
        string = string[:-2]
        return string


def dicts_to_dataclasses(cls):
    if issubclass(cls, DataClass):
        init = cls.__init__

        def new_init(self, **kwargs):
            init(self, **kwargs)
            self.__ad_to_dt__()

        cls.__init__ = new_init
    else:
        raise TypeError("Задекорированный класс должен наследоваться от класса DataClass")


def call_method(method_name: str, objects: tuple, *args, **kwargs):
    for obj in objects:
        getattr(obj, method_name)(*args, **kwargs)


if platform.system() == 'Windows':
    try:
        from tkinter import *

        class TkTemplate(Tk):
            def __init__(self, title: str, width: int, height: int, background: str = 'white', resizable: bool = False, **kwargs):
                super().__init__()
                self.title(title)
                self.resizable(resizable, resizable)
                self.geometry(f'{width}x{height}')
                self['bg'] = background
                if if_dict_key(kwargs, 'icon'):
                    self.iconbitmap(kwargs['icon'])
                self.stylesheet = dict_keyf(kwargs, 'stylesheet', dict())
                self.__stylesheet__()
                self.widgets_init()

            def widgets_init(self):
                pass

            def __stylesheet__(self):
                self.stylesheet_class = DataClass(**self.stylesheet)
    except (ModuleNotFoundError, ImportError):
        class TkTemplate:
            def __new__(cls, *args, **kwargs):
                raise OSError("No module named 'tkinter'")

            def __init__(self, *args, **kwargs):
                raise OSError("No module named 'tkinter'")
else:
    class TkTemplate:
        def __new__(cls, *args, **kwargs):
            raise OSError('This class is only supported on Windows')

        def __init__(self, *args, **kwargs):
            raise OSError('This class is only supported on Windows')


class Color:
    def __init__(self, red: int, green: int, blue: int) -> None:
        self.red, self.green, self.blue = red, green, blue
        self.hexCode = self.__hex__()

    def __hex__(self) -> str:
        return '#' + dec_to_hex(self.red) + dec_to_hex(self.green) + dec_to_hex(self.blue)

    def __dict__(self) -> dict:
        return {'red': self.red, 'green': self.green, 'blue': self.blue, 'hex': self.hexCode}

    def shuffle(self, hexCode: str) -> None:
        self.red = (self.red + hex_to_dec(hexCode[1:3])) // 2
        self.green = (self.green + hex_to_dec(hexCode[3:5])) // 2
        self.blue = (self.blue + hex_to_dec(hexCode[5:7])) // 2
        self.hexCode = self.__hex__()


class Class:
    """Обычный пустой класс"""


class range_plus:
    def __init__(self, *args, **kwargs):
        if kwargs:
            new_args = (kwargs.get('start', 1), kwargs['end'], kwargs.get('step', 1))
        elif args:
            args = list(args)
            largs = len(args)
            if largs == 1:
                new_args = (1, args[0] + 1, 1)
            else:
                args[1] += 1
                if largs == 2:
                    new_args = (args[0], args[1], 1)
                elif largs == 3:
                    new_args = args
                else:
                    raise ValueError("Максимум 3 аргумента")
        else:
            raise ValueError("Нужен хотя бы 1 аргумент")
        self.range = range(*new_args)

    def __iter__(self):
        return iter(self.range)

    def __str__(self):
        return str(self.range)


def emptydataclass(cls):
    def str__(self):
        return super(type(self), self).__str__().replace(self.__class__.__name__, '', 1)
    cls = dataclass(cls)
    cls.__str__ = str__
    return cls


class RawString(str):
    def __add__(self, other):
        if not isinstance(other, str):
            other = str(other)
        return RawString(super().__add__(other))

    def __radd__(self, other):
        return RawString(str(other) + str(self))  # Гарантированно работает для любых типов

    def convert(self):
        return str(self)

    def __eq__(self, other):
        return other == self or str(other) == str(self)


class NoneType:
    def __bool__(self):
        return False

    def __int__(self):
        return 0

    def __float__(self):
        return 0.0

    def __str__(self):
        return ''

    def __eq__(self, other):
        return other is None or other == self

    def __ne__(self, other):
        return not self.__eq__(other)


class frozendict(dict):
    def __setitem__(self, key, value):
        raise TypeError(f"'{self.__class__.__name__}' object does not support item assignment")

    def __delitem__(self, key):
        raise TypeError(f"'{self.__class__.__name__}' object does not support item deletion")


class TransposedList:
    def __init__(self, data):
        """
        Инициализирует объект с данными для транспонирования.

        Args:
            data: Итерируемый объект с вложенными итерируемыми объектами одинаковой длины
        """
        self._validate_data(data)
        self._data = data

    def _validate_data(self, data):
        """Проверяет, что данные можно транспонировать."""
        try:
            iter(data)  # Проверяем, что объект итерируемый
            if not data:
                return  # Пустые данные допустимы

            # Проверяем, что все вложенные элементы имеют одинаковую длину
            first_len = len(data[0]) if hasattr(data[0], '__len__') else len(list(data[0]))
            for item in data:
                current_len = len(item) if hasattr(item, '__len__') else len(list(item))
                if current_len != first_len:
                    raise ValueError("All sub-iterables must have the same length")
        except TypeError as e:
            raise TypeError("Input data must be iterable") from e
        except IndexError as e:
            raise ValueError("Input data cannot be empty") from e

    def raw(self):
        """Возвращает исходные данные в виде списка."""
        return list(self._data) if not hasattr(self._data, '__len__') else self._data

    def __len__(self):
        """Возвращает количество строк в транспонированном представлении."""
        if not self._data:
            return 0
        first_item = self._data[0]
        return len(first_item) if hasattr(first_item, '__len__') else len(list(first_item))

    def __iter__(self):
        """Возвращает итератор по транспонированным данным."""
        # Если данные пустые, возвращаем пустой итератор
        if not self._data:
            return iter([])

        # Создаем итераторы для всех вложенных последовательностей
        iterators = [iter(subseq) for subseq in self._data]

        # Генерируем транспонированные строки
        while True:
            try:
                # Собираем элементы из каждого итератора
                yield [next(it) for it in iterators]
            except StopIteration:
                break

    def __getitem__(self, index):
        """Возвращает транспонированную строку по индексу."""
        try:
            # Проверяем, что индекс допустим
            if not isinstance(index, (int, slice)):
                raise TypeError("Index must be integer or slice")

            # Если данные пустые, вызываем исключение
            if not self._data:
                raise IndexError("Index out of range")

            # Получаем длину первой подпоследовательности
            first_len = len(self._data[0]) if hasattr(self._data[0], '__len__') else len(list(self._data[0]))

            # Обработка целочисленного индекса
            if isinstance(index, int):
                if index < -first_len or index >= first_len:
                    raise IndexError("Index out of range")

                # Возвращаем транспонированную строку
                return [subseq[index] if hasattr(subseq, '__getitem__') else list(subseq)[index]
                        for subseq in self._data]

            # Обработка слайса
            else:
                # Преобразуем в список, чтобы можно было делать несколько итераций
                data_list = list(self._data)
                # Получаем транспонированный список
                transposed = list(zip(*data_list))
                # Применяем слайс
                sliced = transposed[index]
                # Преобразуем обратно в список списков (а не кортежей)
                return [list(row) for row in sliced]

        except (IndexError, TypeError) as e:
            raise type(e)(f"Failed to get item at index {index}: {str(e)}") from e

    def __setitem__(self, index, value):
        """Устанавливает значение в транспонированной позиции."""
        try:
            # Проверяем, что индекс допустим
            if not isinstance(index, int):
                raise TypeError("Index must be integer for assignment")

            # Проверяем, что данные не пустые
            if not self._data:
                raise IndexError("Cannot assign to empty transposed sequence")

            # Проверяем, что значение имеет правильную длину
            if len(value) != len(self._data):
                raise ValueError(f"Value must have length {len(self._data)}")

            # Устанавливаем значения в исходные последовательности
            for i, subseq in enumerate(self._data):
                # Проверяем, поддерживает ли подпоследовательность присваивание
                if hasattr(subseq, '__setitem__'):
                    subseq[index] = value[i]
                else:
                    raise TypeError(f"Subsequence at position {i} does not support item assignment")

        except (IndexError, TypeError, ValueError) as e:
            raise type(e)(f"Failed to set item at index {index}: {str(e)}") from e

    def __delitem__(self, index):
        """Удаляет транспонированную строку по индексу."""
        try:
            # Проверяем, что индекс допустим
            if not isinstance(index, int):
                raise TypeError("Index must be integer for deletion")

            # Проверяем, что данные не пустые
            if not self._data:
                raise IndexError("Cannot delete from empty transposed sequence")

            # Удаляем элементы из исходных последовательностей
            for subseq in self._data:
                # Проверяем, поддерживает ли подпоследовательность удаление
                if hasattr(subseq, '__delitem__'):
                    del subseq[index]
                else:
                    raise TypeError("Subsequence does not support item deletion")

        except (IndexError, TypeError) as e:
            raise type(e)(f"Failed to delete item at index {index}: {str(e)}") from e

    def __repr__(self):
        """Строковое представление объекта."""
        return f"TransposedList({self._data})"

    def __str__(self):
        """Строковое представление транспонированных данных."""
        transposed = list(self)
        return str(transposed)


class TupleDict:
    def __init__(self, *args: tuple[Any, Any]):
        self._data = list(args)

    def to_dict(self) -> dict:
        output = dict()
        for key, value in self._data:
            output[key] = value
        return output

    def __iter__(self):
        return iter(self.to_dict())

    def __getitem__(self, index):
        return self.to_dict()[index]

    def __setitem__(self, index, value):
        self.to_dict()[index] = value

    def __delitem__(self, index):
        del self.to_dict()[index]



from hrenpack.listwork import dict_keyf


def kwarg_return(kwargs: dict, key: str, false):
    return dict_keyf(kwargs, key, false)


def kwarg_function(kwargs: dict, key: str, true, false):
    if key in kwargs:
        true()
    else:
        false()


def kwarg_kwargs(**kwargs):
    return kwargs


def get_kwarg(kwargs: dict, key: str, default=None, raise_error: bool = True, delete: bool = False):
    if default:
        raise_error = False
    if raise_error and key not in kwargs:
        raise KeyError(key)
    output = kwargs.get(key, default)
    if delete and key in kwargs:
        del kwargs[key]
    return output



import hashlib, bcrypt
from typing import Literal


def hash_password(password: str, sha_alg: Literal[1, 256, 384, 512, 224] = 256) -> str:
    def coding(code):
        return f'hashlib.sha{code}()'

    password_bytes = password.encode('utf-8')
    hashed = eval(coding(sha_alg))
    hashed.update(password_bytes)
    hashed_password = hashed.hexdigest()
    return hashed_password


def hash_password_bcrypt(password: str) -> str:
    password_bytes = password.encode('utf-8')
    hashed_password = bcrypt.hashpw(password_bytes, bcrypt.gensalt())
    return hashed_password.decode('utf-8')


def check_password_bcrypt(password: str, hashed_password: str) -> bool:
    password_bytes = password.encode('utf-8')
    hashed_password_bytes = hashed_password.encode('utf-8')
    return bcrypt.checkpw(password_bytes, hashed_password_bytes)



from hrenpack.decorators import confirm
from hrenpack.filework import TextFile


class CompileError(Exception):
    pass


class ProgrammingLanguageSourceFile(TextFile):
    def __init__(self, path: str, compiler_url: str, options: str):
        super().__init__(path)
        self.compiler_url = compiler_url
        self.options = options

    def online_compile(self, output_filename):
        from requests import post

        data = {'src': self.read(), 'options': self.options}
        response = post(self.compiler_url, data)

        if response.status_code == 200:
            with open(output_filename, 'wb') as file:
                file.write(response.content)
        else:
            raise CompileError("Ошибка при компиляции")


class PythonFile(ProgrammingLanguageSourceFile):
    def __init__(self, path: str):
        super().__init__(path, '', '')
        del self.compiler_url, self.options, self.online_compile

    def run(self, confirm_enabled: bool = True):
        def cd():
            exec(self.read())

        @confirm
        def ce():
            exec(self.read())

        ce() if confirm_enabled else cd()



class SlotsReadMixin:
    def __getitem__(self, item):
        if item in self.__slots__:
            return getattr(self, item)
        raise KeyError(item)


class SlotsMixin(SlotsReadMixin):
    def __setitem__(self, key, value):
        setattr(self, key, value)


class LogMixin:
    def log(self, *values, sep: str = ' ', end: str = '\n', file=None, flush: bool = False):
        if self.log_mode:
            print(*values, sep=sep, end=end, file=file, flush=flush)


class LogPlusMixin(LogMixin):
    log: bool = False



import os, tempfile, uuid, requests
from urllib.parse import urlencode
from tqdm import tqdm
from hrenpack import NullStr


class NetworkError(Exception):
    pass


def connection_check():
    error = NetworkError("Подключение к интернету отсутствует")
    try:
        response = requests.get("https://google.com", timeout=5)
        if response.status_code != 200:
            raise error
        return True
    except requests.ConnectionError:
        raise error


def is_connected() -> bool:
    try:
        connection_check()
    except NetworkError:
        return False
    else:
        return True


def connect_to_site(url: str, **kwargs) -> bool:
    if is_connected():
        response = requests.get(url, **kwargs)
        return response.status_code == 200
    else:
        return False


class TestResponse:
    def __init__(self, response: requests.Response):
        self.response = response

    def __call__(self, *args, **kwargs):
        if self.response.status_code == 200:
            self.success(*args, **kwargs)
        else:
            self.error(*args, **kwargs)

    def success(self, *args, **kwargs):
        pass

    def error(self, *args, **kwargs):
        pass


def download_file(url, name: str, to: str = '', params=None, use_progressbar: bool = False, **request_kwargs):
    path = os.path.join(to, name)
    response = requests.get(url, params, **request_kwargs)
    if use_progressbar:
        total_size = int(response.headers.get('content-length', 0))
        with open(path, 'wb') as file, tqdm(
                desc=name,
                total=total_size,
                unit='B',
                unit_scale=True,
                unit_divisor=1024,
        ) as bar:
            for data in response.iter_content(chunk_size=1024):
                bar.update(len(data))
                file.write(data)
    else:
        with open(path, 'wb') as file:
            file.write(response.content)
    return path


def download_file_to_temp(url, extension: str = '', name: str = '', params=None, use_progressbar: bool = False,
                          **request_kwargs):
    tempdir = tempfile.gettempdir()
    if not name:
        name = str(uuid.uuid4())
    if extension:
        extension = '.' + extension
    path = download_file(url, name + extension, tempdir, params, use_progressbar, **request_kwargs)
    return path


class ReCaptchaV3:
    def __init__(self, secret_key: str, site_key: NullStr, score_threshold: float = 0.5):
        self.secret_key = secret_key
        self.score_threshold = score_threshold
        self.site_key = site_key

    @property
    def js_api_url(self) -> str:
        return f'https://www.google.com/recaptcha/api.js?render={self.site_key}'

    def verify(self, token: str) -> bool:
        if not token:
            return False
        json = requests.post('https://www.google.com/recaptcha/api/siteverify',
                             {'secret': self.secret_key, 'response': token}).json()
        return json.get('success', False) and float(json.get('score', 0)) >= self.score_threshold


def GET_url(base_url: str, **params) -> str:
    return f'{base_url}?{urlencode(params)}'



from typing import Optional
from PySide6.QtGui import *
from PySide6.QtWidgets import *
from hrenpack import ColorTyping


class BackgroundRubberBand(QRubberBand):
    def __init__(self, shape, parent=None,
                 background: Optional[ColorTyping] = None,
                 frameground: Optional[ColorTyping] = None):
        super().__init__(shape, parent)
        self.background = background
        self.frameground = frameground

    def paintEvent(self, event):
        painter = QPainter(self)
        background_color = QColor(*self.background)
        painter.fillRect(self.rect(), background_color)

        border_color = QColor(*self.frameground)
        painter.setPen(border_color)
        painter.drawRect(self.rect())



import os
from flask import Blueprint, send_from_directory
from jinja2 import FileSystemLoader, ChoiceLoader

from hrenpack.typings import PathLike


class MultiTemplateAndStaticMixin:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._template_folders = list()
        self._static_folders = list()
        self._static_folders_blueprints = list()

    def add_template_folder(self, folder_path: PathLike):
        """Добавляет папку с шаблонами"""
        if os.path.exists(folder_path) and folder_path not in self._template_folders:
            self._template_folders.append(str(folder_path))
            self._update_template_loader()

    def add_static_folder(self, name: str, folder_path: PathLike):
        """Добавляет папку со статическими файлами"""
        if os.path.exists(folder_path) and folder_path not in self._static_folders:
            self._static_folders.append(str(folder_path))
            blueprint = Blueprint(name, self.import_name, static_folder=folder_path, static_url_path=f'/static/{name}')
            self._static_folders_blueprints.append(blueprint)
            self.register_blueprint(blueprint)

    def serve_static(self, filename):
        """Обработчик для статических файлов из всех папок"""
        # Ищем файл в дополнительных папках
        for static_folder in self._static_folders:
            file_path = os.path.join(static_folder, filename)
            if os.path.exists(file_path):
                directory = os.path.dirname(file_path)
                actual_filename = os.path.basename(file_path)
                return send_from_directory(directory, actual_filename)

        # Если не нашли в дополнительных папках, используем стандартную логику Flask
        return super().send_static_file(filename)

    def _update_template_loader(self):
        """Обновляет загрузчик шаблонов с учетом всех папок"""
        all_template_folders = [self.template_folder] + self._template_folders
        existing_folders = [f for f in all_template_folders if f and os.path.exists(f)]

        loaders = [FileSystemLoader(folder) for folder in existing_folders]
        if hasattr(self, 'jinja_loader') and self.jinja_loader:
            loaders.append(self.jinja_loader)

        self.jinja_loader = ChoiceLoader(loaders)
        


import os
import sys
from glob import glob
from typing import Union, Optional
from PySide6.QtWidgets import *
from PySide6.QtGui import *
from PySide6.QtCore import *
from pathlike_typing import PathLike
from hrenpack.listwork import split_list
from hrenpack.framework.Pyside6.variables import file_dialog_templates


def messagebox(title: str, message: str, icon: Union[QIcon, str, None] = None):
    msg = QMessageBox()
    msg.setWindowTitle(title)
    msg.setText(message)

    if icon is not None:
        icon = icon if type(icon) is QIcon else QIcon(icon)
        msg.setWindowIcon(icon)

    msg.exec()


def remove_text_and_stylesheet(widget: QWidget):
    widget.setStyleSheet("")
    try:
        widget.setText("")
    except AttributeError:
        pass


def objects_enable(*objects: QObject):
    for o in objects:
        o.setEnabled(True)


def objects_disable(*objects: QObject):
    for o in objects:
        o.setEnabled(False)


def exit(window: QWidget) -> None:
    window.close()
    sys.exit(0)


def file_dialog_read(parent, *file_types: str, title: str = "Выберите файл", directory: str = '/',
                     selected_filter: int = 0, all_files: bool = True) -> Optional[str]:
    file_types = list(file_types)
    if all_files:
        file_types.append(file_dialog_templates.all)
    dialog = QFileDialog()
    options = dialog.options()
    filters = split_list(file_types, ';;')
    sf = file_types[selected_filter]
    filename = dialog.getOpenFileName(parent, title, directory, filters, sf, options)[0]
    return filename if filename else None


def file_dialog_save(parent, *file_types: str, title: str = "Выберите файл", directory: str = '/',
                     selected_filter: int = 0, all_files: bool = True) -> Optional[str]:
    file_types = list(file_types)
    if all_files:
        file_types.append(file_dialog_templates.all)
    dialog = QFileDialog()
    options = dialog.options()
    filters = split_list(file_types, ';;')
    sf = file_types[selected_filter]
    filename = dialog.getSaveFileName(parent, title, directory, filters, sf, options)[0]
    return filename


def add_fonts(directory: PathLike, root_only: bool = False):
    fonts = glob(f'{directory}/*.otf') + glob(f'{directory}/*.ttf')
    if not root_only:
        fonts += glob(f'{directory}/static/*.otf') + glob(f'{directory}/static/*.ttf')
    for font in fonts:
        QFontDatabase.addApplicationFont(font)



from kivy.uix.screenmanager import Screen as KivyScreen, ScreenManager as KivyScreenManager, ScreenManagerException


class ScreenManager(KivyScreenManager):
    screens_dict = dict()

    def setCurrentScreen(self, name: str):
        self.current = name

    def getCurrentScreen(self):
        return self.screens_dict[self.current]

    def add_screen(self, screen, name: str, **kwargs):
        screen.manager = self
        screen.name = name
        if kwargs.get('add_layout_to_app', False):
            exec(f'screen.app.screen_{name}Layout = screen.layout')
        self.screens.append(screen)
        self.screens_dict[name] = screen
        if self.current is None:
            self.current = screen.name


class Screen(KivyScreen):
    def __init__(self, manager=None, name: str = '', **kwargs):
        super().__init__(**kwargs)
        if manager is not None and name != '':
            manager.add_widget(self, name=name)
        elif manager is not None:
            manager.add_widget(self)

    def add(self, *widgets):
        for widget in widgets:
            self.add_widget(widget)



import os, sys, locale


def get_system_language() -> str:
    env_lang = os.environ.get('LANG') or os.environ.get('LANGUAGE')
    sys_lang = None
    default_lang = None

    if sys.platform == 'win32':
        try:
            import ctypes
            windll = ctypes.windll.kernel32
            sys_lang = locale.windows_locale[windll.GetUserDefaultUILanguage()].split('_')[0]
        except:
            pass

    try:
        default_lang = locale.getdefaultlocale()[0].split('_')[0]
    except:
        pass

    return env_lang or sys_lang or default_lang or 'en'



from hrenpack.classes import DataClass


class ColorsANSI(DataClass):
    BLACK = '\033[30m'
    RED = '\033[31m'
    GREEN = '\033[32m'
    YELLOW = '\033[33m'
    BLUE = '\033[34m'
    MAGENTA = '\033[35m'
    CYAN = '\033[36m'
    WHITE = '\033[37m'
    RESET = '\033[0m'
    DICT = {
        'black': BLACK, 'red': RED, 'green': GREEN, 'yellow': YELLOW, 'blue': BLUE,
        'magenta': MAGENTA, 'cyan': CYAN, 'white': WHITE, 'reset': RESET,
        "черный": BLACK, "красный": RED, "зеленый": GREEN, "желтый": YELLOW, "синий": BLUE,
        "пурпурный": MAGENTA, "бирюзовый": CYAN, "белый": WHITE, "сброс": RESET
    }


def print_color(*values, color: str, separator: str = ' ', end: str = '\n', file=None) -> None:
    text = separator.join(values)
    text = ColorsANSI.DICT[color] + text + '\033[0m'
    print(text, end=end, file=file)


def print_error(*values, separator: str = ' ', end: str = '\n', file=None) -> None:
    print_color(*values, color='red', separator=separator, end=end, file=file)


def print_success(*values, separator: str = ' ', end: str = '\n', file=None) -> None:
    print_color(*values, color='green', separator=separator, end=end, file=file)


def print_warning(*values, separator: str = ' ', end: str = '\n', file=None) -> None:
    print_color(*values, color='yellow', separator=separator, end=end, file=file)


def print_info(*values, separator: str = ' ', end: str = '\n', file=None) -> None:
    print_color(*values, color='blue', separator=separator, end=end, file=file)


def color_format(*values, color: str, separator: str = ' ', end: str = '') -> str:
    text = separator.join(values)
    text = ColorsANSI.DICT[color] + text + '\033[0m'
    text += end
    return text


def error_format(*values, separator: str = ' ', end: str = '') -> str:
    return color_format(*values, color='red', separator=separator, end=end)


def success_format(*values, separator: str = ' ', end: str = '') -> str:
    return color_format(*values, color='green', separator=separator, end=end)


def warning_format(*values, separator: str = ' ', end: str = '') -> str:
    return color_format(*values, color='yellow', separator=separator, end=end)


def info_format(*values, separator: str = ' ', end: str = '') -> str:
    return color_format(*values, color='blue', separator=separator, end=end)



import logging
from contextlib import redirect_stdout, nullcontext
from functools import wraps
from hrenpack.listwork import key_in_dict


def confirm(inp_text: str = "Вы уверены, что хотите выполнить эту программу?"):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            result = input(inp_text + "\nВведите y, д или 1, если да, или n, н или 0, если нет\n")
            while True:
                if result in ('y', 'Y', "д", "Д", "1"):
                    return func(*args, **kwargs)
                elif result in ('n', 'N', "н", "Н", "0"):
                    break
                else:
                    result = input(inp_text + "\nВведите y, д или 1, если да, или n, н или 0, если нет\n")

        return wrapper
    return decorator


def non_print(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        with redirect_stdout(nullcontext()):
            return func(*args, **kwargs)
    return wrapper


def args_kwargs(**kwargs):
    args_name = kwargs.get('args_name', 'args')
    kwargs_name = kwargs.get('kwargs_name', 'kwargs')
    copy_args = kwargs.get('copy_args', True)
    copy_kwargs = kwargs.get('copy_kwargs', True)
    del_kwargs = kwargs.get('del_kwargs', True)

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **key_args):
            if key_in_dict(key_args, args_name) and copy_args:
                args = [*args, *key_args[args_name]]
                if del_kwargs:
                    del key_args[args_name]
            if key_in_dict(key_args, kwargs_name) and copy_kwargs:
                key_args = {**key_args, **key_args[kwargs_name]}
                if del_kwargs:
                    del key_args[kwargs_name]
            return func(*args, **key_args)
        return wrapper
    return decorator


def debug_logging(start_message: str = '', end_message: str = ''):
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if start_message:
                logging.debug(start_message)
            output = func(*args, **kwargs)
            if end_message:
                logging.debug(end_message)
            return output
        return wrapper
    return decorator


def method(func):
    return func


def multi_decorator(*decorators):
    """Декораторы применяются слева направо"""
    def decorator(func):
        for dec in decorators:
            func = dec(func)
        return func
    return decorator



MIME_TYPES = {
    # Текстовые форматы
    "txt": "text/plain",
    "html": "text/html",
    "htm": "text/html",
    "shtml": "text/html",
    "css": "text/css",
    "csv": "text/csv",
    "ics": "text/calendar",
    "ifb": "text/calendar",
    "md": "text/markdown",
    "markdown": "text/markdown",
    "rtx": "text/richtext",
    "rtf": "text/rtf",
    "sgml": "text/sgml",
    "sgm": "text/sgml",
    "tsv": "text/tab-separated-values",
    "vcf": "text/vcard",
    "vcard": "text/vcard",
    "curl": "text/vnd.curl",
    "dcurl": "text/vnd.curl.dcurl",
    "scurl": "text/vnd.curl.scurl",
    "mcurl": "text/vnd.curl.mcurl",
    "fly": "text/vnd.fly",
    "flx": "text/vnd.fmi.flexstor",
    "gv": "text/vnd.graphviz",
    # "dot": "text/vnd.graphviz",
    "3dml": "text/vnd.in3d.3dml",
    "spot": "text/vnd.in3d.spot",
    "jad": "text/vnd.sun.j2me.app-descriptor",
    "wml": "text/vnd.wap.wml",
    "wmls": "text/vnd.wap.wmlscript",
    "vtt": "text/vtt",
    "s": "text/x-asm",
    "asm": "text/x-asm",
    "c": "text/x-c",
    "h": "text/x-c",
    "cpp": "text/x-c++",
    "cxx": "text/x-c++",
    "cc": "text/x-c++",
    "hpp": "text/x-c++",
    "hh": "text/x-c++",
    "cs": "text/x-csharp",
    "f": "text/x-fortran",
    "for": "text/x-fortran",
    "f77": "text/x-fortran",
    "f90": "text/x-fortran",
    "java": "text/x-java",
    "p": "text/x-pascal",
    "pas": "text/x-pascal",
    "py": "text/x-python",
    "pyw": "text/x-python",
    "scala": "text/x-scala",
    "nfo": "text/x-nfo",
    "opml": "text/x-opml",
    "org": "text/x-org",
    "pod": "text/x-pod",
    "tex": "text/x-tex",
    "ltx": "text/x-tex",
    "sty": "text/x-tex",
    "cls": "text/x-tex",
    "vcs": "text/x-vcalendar",
    "js": "application/javascript",  # Официально с 2022
    "mjs": "application/javascript",

    # Изображения
    "avif": "image/avif",
    "bmp": "image/bmp",
    "dib": "image/bmp",
    "gif": "image/gif",
    "heic": "image/heic",
    "heif": "image/heif",
    "jpg": "image/jpeg",
    "jpeg": "image/jpeg",
    "jpe": "image/jpeg",
    "jif": "image/jpeg",
    "jfif": "image/jpeg",
    "jfi": "image/jpeg",
    "png": "image/png",
    "svg": "image/svg+xml",
    "svgz": "image/svg+xml",
    "tiff": "image/tiff",
    "tif": "image/tiff",
    "webp": "image/webp",
    "ico": "image/x-icon",
    "icon": "image/x-icon",
    "pcx": "image/x-pcx",
    "pct": "image/x-pict",
    "pic": "image/x-pict",
    "pnm": "image/x-portable-anymap",
    "pbm": "image/x-portable-bitmap",
    "pgm": "image/x-portable-graymap",
    "ppm": "image/x-portable-pixmap",
    "rgb": "image/x-rgb",
    "xbm": "image/x-xbitmap",
    "xpm": "image/x-xpixmap",
    "xwd": "image/x-xwindowdump",

    # Аудио
    "aac": "audio/aac",
    "flac": "audio/flac",
    "mid": "audio/midi",
    "midi": "audio/midi",
    "kar": "audio/midi",
    "rmi": "audio/midi",
    "mp3": "audio/mpeg",
    "mpga": "audio/mpeg",
    "mp2": "audio/mpeg",
    "mp2a": "audio/mpeg",
    "m2a": "audio/mpeg",
    "m3a": "audio/mpeg",
    "oga": "audio/ogg",
    "ogg": "audio/ogg",
    "spx": "audio/ogg",
    "opus": "audio/ogg",
    "wav": "audio/wav",
    "weba": "audio/webm",
    "aif": "audio/x-aiff",
    "aiff": "audio/x-aiff",
    "aifc": "audio/x-aiff",
    "m4a": "audio/x-m4a",
    "mka": "audio/x-matroska",
    "wma": "audio/x-ms-wma",
    "ram": "audio/x-pn-realaudio",
    "ra": "audio/x-pn-realaudio",

    # Видео
    "3gp": "video/3gpp",
    "3gpp": "video/3gpp",
    "3g2": "video/3gpp2",
    "h261": "video/h261",
    "h263": "video/h263",
    "h264": "video/h264",
    "jpgv": "video/jpeg",
    "jpm": "video/jpm",
    "jpgm": "video/jpm",
    "mj2": "video/mj2",
    "mjp2": "video/mj2",
    "mp4": "video/mp4",
    "mp4v": "video/mp4",
    "mpg4": "video/mp4",
    "mpeg": "video/mpeg",
    "mpg": "video/mpeg",
    "mpe": "video/mpeg",
    "m1v": "video/mpeg",
    "m2v": "video/mpeg",
    "ogv": "video/ogg",
    "qt": "video/quicktime",
    "mov": "video/quicktime",
    "mxu": "video/vnd.mpegurl",
    "m4u": "video/vnd.mpegurl",
    "webm": "video/webm",
    "f4v": "video/x-f4v",
    "flv": "video/x-flv",
    "m4v": "video/x-m4v",
    "mkv": "video/x-matroska",
    "mk3d": "video/x-matroska",
    "mks": "video/x-matroska",
    "asf": "video/x-ms-asf",
    "asx": "video/x-ms-asf",
    "wm": "video/x-ms-wm",
    "wmv": "video/x-ms-wmv",
    "wmx": "video/x-ms-wmx",
    "wvx": "video/x-ms-wvx",
    "avi": "video/x-msvideo",
    "movie": "video/x-sgi-movie",
    "smv": "video/x-smv",

    # Документы и офисные форматы
    "pdf": "application/pdf",
    "doc": "application/msword",
    "dot": "application/msword",
    "xls": "application/vnd.ms-excel",
    "xlt": "application/vnd.ms-excel",
    "xla": "application/vnd.ms-excel",
    "ppt": "application/vnd.ms-powerpoint",
    "pot": "application/vnd.ms-powerpoint",
    "pps": "application/vnd.ms-powerpoint",
    "ppa": "application/vnd.ms-powerpoint",
    "docx": "application/vnd.openxmlformats-officedocument.wordprocessingml.document",
    "dotx": "application/vnd.openxmlformats-officedocument.wordprocessingml.template",
    "xlsx": "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
    "xltx": "application/vnd.openxmlformats-officedocument.spreadsheetml.template",
    "pptx": "application/vnd.openxmlformats-officedocument.presentationml.presentation",
    "potx": "application/vnd.openxmlformats-officedocument.presentationml.template",
    "ppsx": "application/vnd.openxmlformats-officedocument.presentationml.slideshow",
    "odt": "application/vnd.oasis.opendocument.text",
    "ods": "application/vnd.oasis.opendocument.spreadsheet",
    "odp": "application/vnd.oasis.opendocument.presentation",
    "odg": "application/vnd.oasis.opendocument.graphics",
    "odc": "application/vnd.oasis.opendocument.chart",
    "odf": "application/vnd.oasis.opendocument.formula",
    "odb": "application/vnd.oasis.opendocument.database",
    "odi": "application/vnd.oasis.opendocument.image",
    "odm": "application/vnd.oasis.opendocument.text-master",
    "epub": "application/epub+zip",

    # Архивы
    "gz": "application/gzip",
    "gzip": "application/gzip",
    "bz": "application/x-bzip",
    "bz2": "application/x-bzip2",
    "boz": "application/x-bzip2",
    "7z": "application/x-7z-compressed",
    "rar": "application/x-rar-compressed",
    "tar": "application/x-tar",
    "zip": "application/zip",
    "ace": "application/x-ace-compressed",
    "afa": "application/x-astrotite-afa",
    "alz": "application/x-alz-compressed",
    "dmg": "application/x-apple-diskimage",
    "arj": "application/x-arj",
    "cfs": "application/x-cfs-compressed",
    "dgc": "application/x-dgc-compressed",
    "lzh": "application/x-lzh",
    "lha": "application/x-lzh",
    "lzx": "application/x-lzx",
    "rpm": "application/x-rpm",
    "sit": "application/x-stuffit",
    "sitx": "application/x-stuffitx",
    "xz": "application/x-xz",

    # Данные и форматы обмена
    "json": "application/json",
    "jsonld": "application/ld+json",
    "xml": "application/xml",
    "xhtml": "application/xhtml+xml",
    "xht": "application/xhtml+xml",
    "xsl": "application/xslt+xml",
    "xslt": "application/xslt+xml",
    "php": "application/x-httpd-php",
    "phtml": "application/x-httpd-php",
    "php3": "application/x-httpd-php",
    "php4": "application/x-httpd-php",
    "php5": "application/x-httpd-php",
    "phps": "application/x-httpd-php-source",
    "es": "application/ecmascript",
    "ecma": "application/ecmascript",
    "wasm": "application/wasm",
    "sql": "application/sql",
    "cgi": "application/x-httpd-cgi",
    "sh": "application/x-sh",
    "csh": "application/x-csh",

    # Веб и интернет
    "atom": "application/atom+xml",
    "rss": "application/rss+xml",
    "jar": "application/x-java-archive",
    "jnlp": "application/x-java-jnlp-file",
    "run": "application/x-makeself",
    "pl": "application/x-perl",
    "pm": "application/x-perl",
    # "prc": "application/x-pilot",
    # "pdb": "application/x-pilot",
    "sea": "application/x-sea",
    "swf": "application/x-shockwave-flash",
    "xap": "application/x-silverlight-app",

    # Бинарные и исполняемые файлы
    "bin": "application/octet-stream",
    "dms": "application/octet-stream",
    "lrf": "application/octet-stream",
    "mar": "application/octet-stream",
    "so": "application/octet-stream",
    "dist": "application/octet-stream",
    "distz": "application/octet-stream",
    "pkg": "application/octet-stream",
    "bpk": "application/octet-stream",
    "dump": "application/octet-stream",
    "elc": "application/octet-stream",
    "deploy": "application/octet-stream",
    "exe": "application/x-msdownload",
    "dll": "application/x-msdownload",
    "com": "application/x-msdownload",
    "bat": "application/x-msdownload",
    "msi": "application/x-msdownload",
    "deb": "application/x-debian-package",
    "udeb": "application/x-debian-package",
    "iso": "application/x-iso9660-image",
    "img": "application/octet-stream",
    "msp": "application/octet-stream",
    "msm": "application/octet-stream",

    # Криптография и сертификаты
    "pgp": "application/pgp-encrypted",
    "sig": "application/pgp-signature",
    "asc": "application/pgp-signature",
    "p10": "application/pkcs10",
    "p7m": "application/pkcs7-mime",
    "p7c": "application/pkcs7-mime",
    "p7s": "application/pkcs7-signature",
    "p8": "application/pkcs8",
    "ac": "application/pkix-attr-cert",
    "cer": "application/pkix-cert",
    "crl": "application/pkix-crl",
    "pkipath": "application/pkix-pkipath",
    "pki": "application/pkixcmp",

    # Графика и векторные форматы
    "ps": "application/postscript",
    "ai": "application/postscript",
    "eps": "application/postscript",

    # Шрифты
    "ttc": "font/collection",
    "otf": "font/otf",
    "ttf": "font/ttf",
    "woff": "font/woff",
    "woff2": "font/woff2",

    # 3D модели
    "3mf": "model/3mf",
    "gltf": "model/gltf+json",
    "glb": "model/gltf-binary",
    "igs": "model/iges",
    "iges": "model/iges",
    "msh": "model/mesh",
    "mesh": "model/mesh",
    "silo": "model/mesh",
    "stl": "model/stl",
    "wrl": "model/vrml",
    "vrml": "model/vrml",
    "x3db": "model/x3d+binary",
    "x3dbz": "model/x3d+binary",
    "x3dv": "model/x3d+vrml",
    "x3dvz": "model/x3d+vrml",
    "x3d": "model/x3d+xml",
    "x3dz": "model/x3d+xml",

    # Химические форматы
    "cdx": "chemical/x-cdx",
    "cif": "chemical/x-cif",
    "cmdf": "chemical/x-cmdf",
    "cml": "chemical/x-cml",
    "csml": "chemical/x-csml",
    "pdb": "chemical/x-pdb",
    "xyz": "chemical/x-pdb",

    # Электронные книги
    "azw": "application/vnd.amazon.ebook",
    "prc": "application/x-mobipocket-ebook",
    "mobi": "application/x-mobipocket-ebook",

    # Форматы данных ГИС и карт
    "kml": "application/vnd.google-earth.kml+xml",
    "kmz": "application/vnd.google-earth.kmz",
    "gpx": "application/gpx+xml",
    "osm": "application/vnd.openstreetmap.data+xml",

    # Специализированные форматы приложений
    "apk": "application/vnd.android.package-archive",
    "m3u8": "application/vnd.apple.mpegurl",
    "mpkg": "application/vnd.apple.installer+xml",
    "xul": "application/vnd.mozilla.xul+xml",
    "xpi": "application/x-xpinstall",
    "crx": "application/x-chrome-extension",
    "eot": "application/vnd.ms-fontobject",
    "chm": "application/vnd.ms-htmlhelp",
    "ims": "application/vnd.ms-ims",
    "lrm": "application/vnd.ms-lrm",
    "thmx": "application/vnd.ms-officetheme",
    "msg": "application/vnd.ms-outlook",
    "cat": "application/vnd.ms-pki.seccat",
    "xps": "application/vnd.ms-xpsdocument",
    "mdb": "application/x-msaccess",
    "obd": "application/x-msbinder",
    "crd": "application/x-mscardfile",
    "clp": "application/x-msclip",
    "mvb": "application/x-msmediaview",
    "m13": "application/x-msmediaview",
    "m14": "application/x-msmediaview",
    "wmf": "application/x-msmetafile",
    "emf": "application/x-msmetafile",
    "emz": "application/x-msmetafile",
    "mny": "application/x-msmoney",
    "pub": "application/x-mspublisher",
    "scd": "application/x-msschedule",
    "trm": "application/x-msterminal",
    "wri": "application/x-mswrite",
    "nc": "application/x-netcdf",
    "cdf": "application/x-netcdf",
    "nzb": "application/x-nzb",
    "p12": "application/x-pkcs12",
    "pfx": "application/x-pkcs12",
    "p7b": "application/x-pkcs7-certificates",
    "spc": "application/x-pkcs7-certificates",
    "p7r": "application/x-pkcs7-certreqresp",
    "ris": "application/x-research-info-systems",
    "shar": "application/x-shar",
    "srt": "application/x-subrip",
    "sv4cpio": "application/x-sv4cpio",
    "sv4crc": "application/x-sv4crc",
    "t3": "application/x-t3vm-image",
    "gam": "application/x-tads",
    "tcl": "application/x-tcl",
    "tk": "application/x-tcl",
    "texinfo": "application/x-texinfo",
    "texi": "application/x-texinfo",
    "obj": "application/x-tgif",
    "ustar": "application/x-ustar",
    "src": "application/x-wais-source",
    "der": "application/x-x509-ca-cert",
    "crt": "application/x-x509-ca-cert",
    "pem": "application/x-x509-ca-cert",
    "fig": "application/x-xfig",
    "xlf": "application/x-xliff+xml",
    "z1": "application/x-zmachine",
    "z2": "application/x-zmachine",
    "z3": "application/x-zmachine",
    "z4": "application/x-zmachine",
    "z5": "application/x-zmachine",
    "z6": "application/x-zmachine",
    "z7": "application/x-zmachine",
    "z8": "application/x-zmachine",
    "xaml": "application/xaml+xml",
    "xav": "application/xcap-att+xml",
    "xca": "application/xcap-caps+xml",
    "xdf": "application/xcap-diff+xml",
    "xel": "application/xcap-el+xml",
    "xns": "application/xcap-ns+xml",
    "xenc": "application/xenc+xml",
    "xop": "application/xop+xml",
    "xpl": "application/xproc+xml",
    "xspf": "application/xspf+xml",
    "mxml": "application/xv+xml",
    "xhvml": "application/xv+xml",
    "xvml": "application/xv+xml",
    "xvm": "application/xv+xml",
    "yang": "application/yang",
    "yin": "application/yin+xml",
    "wgt": "application/widget",
    "hlp": "application/winhlp",

    # Дополнительные форматы
    "abw": "application/x-abiword",
    "arc": "application/x-freearc",
    "spl": "application/x-futuresplash",
    "gca": "application/x-gca-compressed",
    "ulx": "application/x-glulx",
    "gnumeric": "application/x-gnumeric",
    "gramps": "application/x-gramps-xml",
    "gtar": "application/x-gtar",
    "hdf": "application/x-hdf",
    "install": "application/x-install-instructions",
    "jardiff": "application/x-java-archive-diff",
    "latex": "application/x-latex",
    "mie": "application/x-mie",
    "application": "application/x-ms-application",
    "lnk": "application/x-ms-shortcut",
    "wmd": "application/x-ms-wmd",
    "wmz": "application/x-ms-wmz",
    "xbap": "application/x-ms-xbap",

    # Сообщения и почта
    "eml": "message/rfc822",
    "mime": "message/rfc822",

    # Виртуальная реальность
    "usdz": "model/vnd.usdz+zip",
    "vmt": "application/vnd.valve.source.material",
    "vtf": "application/vnd.valve.source.texture",
}

# Дополнительная функция для обработки составных расширений
COMPOUND_EXTENSIONS = {
    "tar.gz": "application/gzip",
    "tar.bz2": "application/x-bzip2",
    "tar.xz": "application/x-xz",
}



import os, json, csv, functools
from typing import Union, Literal, Any, List
from hrenpack import one_return
from hrenpack.cmd import get_filename, get_extension, create_file, delete_file, FileNameInfo
from hrenpack.listwork import (split_list, _is_tuple, list_add, split_list_enter, split_list_space, key_in_dict,
                               equals_keys)
from hrenpack.strwork import search_and_edit
from hrenpack.decorators import confirm
from configparser import ConfigParser, NoOptionError, NoSectionError


class FileTypeError(Exception):
    pass


class FileIsNotEmptyError(Exception):
    pass


class FileIsEmptyError(Exception):
    pass


def extension_check(path: str, *extensions: Union[list, tuple, str]) -> None:
    extensions_converted = list()
    for e in extensions:
        if type(e) is str:
            extensions_converted.append(e)
        else:
            for g in e:
                extensions_converted.append(g)

    extension = get_extension(path)
    if extension not in extensions_converted:
        raise FileTypeError(f"Данный класс предназначен для {split_list(extensions)} файлов. "
                            f"Ваш файл с расширением {extension}")


def create_file_if_not_exists(path: Union[str, FileNameInfo]) -> None:
    path = path if type(path) is str else path.path
    if not os.path.isfile(path):
        create_file(path)


def use_default(func):
    @functools.wraps(func)
    def wrapper(self, section: str, key: str, default: Any = None):
        try:
            return func(self, section, key, default)
        except (NoOptionError, NoSectionError):
            return default
    return wrapper


class TextFile:
    comment_letter = ''

    def __init__(self, path: Union[str, FileNameInfo], encoding: Union[str, int] = 'utf-8', **kwargs):
        self.path = path if type(path) is str else path.path
        self.get_filename = lambda: get_filename(self.path)
        self.get_extension = lambda: get_extension(self.path)
        self.encoding = str(encoding)
        self.search_and_delete = lambda input: self.search_and_edit(input, '')
        self._comments = list()
        if key_in_dict(kwargs, 'extension'):
            extension_check(path, kwargs['extension'])
        elif key_in_dict(kwargs, 'extensions'):
            extension_check(path, *kwargs['extensions'])
        create_file_if_not_exists(self.path)

    @staticmethod
    def comment_decorator(func):
        def wrapper(self, *args, **kwargs):
            if self.comment_letter == '':
                raise AttributeError(
                    f"AttributeError: '{self.__class__.__name__}' object has no attribute 'add_comment'")
            return func(self, *args, **kwargs)
        return wrapper

    def read(self, letters: int = -1) -> str:
        file = open(self.path, encoding=self.encoding)
        data = file.read(letters)
        file.close()
        return data

    def rewrite(self, data: str):
        file = open(self.path, 'w', encoding=self.encoding)
        file.write(data)
        file.close()

    def add_data(self, data: str, separator: str = ''):
        file = open(self.path, 'a', encoding=self.encoding)
        file.write(separator)
        file.write(data)
        file.close()

    @comment_decorator
    def add_comment(self, comment: str):
        if '\n' in comment:
            raise ValueError('\\n in comment')
        self.add_data('{} {}'.format(self.comment_letter, comment))

    def copy(self, new_path: str):
        if not os.path.isfile(new_path):
            raise FileExistsError(
                f'[WinError 183] Невозможно создать новый файл, так как он уже существует: {new_path}'
            )
        else:
            data = self.read()
            file = open(new_path, 'w', encoding=self.encoding)
            file.write(data)
            file.close()

    def copy_and_edit_text(self, new_path: str, new_text: str):
        self.copy(new_path)
        file = TextFile(new_path)
        file.rewrite(new_text)

    def copy_with_prefix(self, prefix_text: str, is_suffix: bool = False,
                         new_text: Union[str, bool] = False, is_return_filename: bool = False) -> str:
        if is_suffix:
            new_filename = f'{self.path} {prefix_text}'
        else:
            new_filename = f'{prefix_text} {self.path}'

        self.copy(new_filename)

        if new_text is not False:
            copyed_file = TextFile(new_filename)
            copyed_file.rewrite(new_text)

        if is_return_filename:
            return new_filename

    # Номер строки (line) считается начиная с единицы
    def edit_line(self, line: int, new_data: str):
        lines = self.read().split('\n')
        lines[line - 1] = new_data
        self.rewrite(split_list(lines, '\n'))

    def is_empty(self) -> bool:
        return self.read() == ''

    def search_and_edit(self, input: str, output: str) -> None:
        new_text = search_and_edit(self.read(), input, output)
        self.rewrite(new_text)

    def write_if_is_empty(self, data):
        if self.is_empty():
            self.rewrite(data)
        else:
            raise FileIsNotEmptyError("Для использования этой функции файл должен быть пустым")

    def read_lines(self, is_tuple: bool = False, without_n: bool = True) -> Union[list[str], tuple[str]]:
        output = self.read().split('\n')
        if not without_n:
            for el in output:
                el += '\n'
        return _is_tuple(output, is_tuple)

    def __copy__(self, new_path: str):
        self.copy(new_path)
        return TextFile(new_path)

    @confirm("Данное действие удалит файл. Вы уверены, что хотите продолжить?")
    def delete(self):
        delete_file(self.path)
        del self

    def newline(self, line: str):
        if self.is_empty():
            self.rewrite(line)
        else:
            self.rewrite(f'{self.read()}\n{line}')

    def __str__(self):
        return self.read()

    def __bool__(self):
        return self.is_empty()

    def open(self,
             mode: Literal[
                 "r+", "+r", "rt+", "r+t", "+rt", "tr+", "t+r", "+tr", "w+", "+w", "wt+", "w+t", "+wt", "tw+", "t+w", "+tw", "a+", "+a", "at+", "a+t", "+at", "ta+", "t+a", "+ta", "x+", "+x", "xt+", "x+t", "+xt", "tx+", "t+x", "+tx", "w", "wt", "tw", "a", "at", "ta", "x", "xt", "tx", "r", "rt", "tr", "U", "rU", "Ur", "rtU", "rUt", "Urt", "trU", "tUr", "Utr"] = "r",
             buffering: int = -1,
             errors: str | None = None,
             newline: str | None = None,
             closefd: bool = True):
        return open(self.path, mode, buffering, self.encoding, errors, newline, closefd)


# Важно! Для использования класса файл SRT должен быть написан по всем правилам (между разными типами данных должен быть
# один переход на новую строку, между секциями должно быть два перехода на новую строку)
class SRTSubtitleFile(TextFile):
    class SubtitleError(Exception):
        pass

    def __init__(self, path: str, encoding: Union[str, int] = 'utf-8'):
        super().__init__(path, encoding)
        extension_check(path, 'srt')
        self.subtitles_timecodes = self.read_subtitle()['timecodes']
        self.subtitles_text = self.read_subtitle()['text']
        self.sections = self.read_subtitle()['number']
        self.edit, self.elst, self.edit_subtitle = one_return(3, self.edit_line_subtitle_text)

    def read_subtitle(self) -> dict:
        try:
            sections = self.read().split('\n\n')
            timecodes, text_data = one_return(2, list())
            for section in sections:
                number, timecode, text = section.split('\n')
                timecodes.append(timecode)
                text_data.append(text)
            return {'timecodes': timecodes, 'text': text_data, 'number': len(sections)}
        except ValueError:
            if not self.is_empty():
                raise self.SubtitleError("Данный файл субтитров построен неправильно. Наша библиотека не может с ним"
                                         " работать. Приносим извинения за предоставленные неудобства.")

    def edit_line_subtitle_text(self, line: int, new_text: str):
        text_line = line * 4 - 1
        self.edit_line(text_line, new_text)

    def edit_timecode(self, line: int, new_timecode_begin: str, new_timecode_end: str):
        text_line = line * 4 - 2
        self.edit_line(text_line, f'{new_timecode_begin} --> {new_timecode_end}')


class ConfigurationFile(TextFile):
    comment_letter = ';'

    class _Section:
        def __init__(self, config: ConfigParser, name: str):
            self.config = dict(config.items(name))

        def __getitem__(self, key: str):
            return self.config[key]

        def __setitem__(self, key: str, value: str):
            self.config[key] = value

        def get(self, key: str, default=None):
            return self.config.get(key, default)

    def __init__(self, path: str = 'config.ini', encoding: Union[str, int] = 'utf-8'):
        super().__init__(path, encoding)
        extension_check(self.path, 'ini')
        self.config = ConfigParser()
        self.read_config = lambda: self.config.read(self.path, encoding=self.encoding)
        self.read_config()
        self.get_bool = self.get_boolean

    @use_default
    def get_value(self, section: str, key: str, default: Any = None) -> str:
        return self.config.get(section, key)

    @use_default
    def get_boolean(self, section: str, key: str, default: Any = False) -> bool:
        return self.config.getboolean(section, key)

    @use_default
    def get_int(self, section: str, key: str, default: Any = 0) -> int:
        return self.config.getint(section, key)

    @use_default
    def get_float(self, section: str, key: str, default: Any = 0) -> float:
        return self.config.getfloat(section, key)

    def set_value(self, section: str, key: str, value) -> None:
        self.config.set(section, key, str(value))
        self.save()

    def save(self):
        with open(self.path, 'w', encoding=self.encoding) as config:
            self.config.write(config)

    def section_exists(self, section: str) -> bool:
        return section in self.config.sections()

    def value_exists(self, section: str, key: str) -> bool:
        return self.config.has_option(section, key)

    def delete_value(self, section: str, key: str) -> None:
        self.config.delete(section, key)
        self.save()

    def delete_section(self, section: str) -> None:
        self.config.delete(section)
        self.save()

    def create_section(self, section: str) -> None:
        self.config.add_section(section)
        self.save()

    def create_key_in_section(self, section: str, key: str, value) -> None:
        if not self.value_exists(section, key):
            if not self.section_exists(section):
                self.create_section(section)
                self.config.set(section, key, str(value))
                self.save()
            else:
                self.config.set(section, key, str(value))
                self.save()
        else:
            if not value == self.get_value(section, key):
                self.set_value(section, key, str(value))

    def add_comment(self, line_index: int, text: str, comment_letter: Literal[';', '#'] = ';'):
        lines = self.read_lines()
        lines = list_add(lines, line_index, split_list_space((comment_letter, text)))
        self.rewrite(split_list_enter(lines))

    def edit_section(self, section: str, block: dict, rewrite: bool = False) -> None:
        if rewrite:
            self.delete_section(section)
            self.create_section(section)
        for key in block:
            self.set_value(section, key, block[key])

    def rewrite(self, data: str):
        with open(self.path, 'w', encoding=self.encoding) as file:
            file.write(data)
            self.read_config()

    def set_bool(self, section: str, key: str, value):
        self.set_value(section, key, bool(value))

    def set_int(self, section: str, key: str, value):
        self.set_value(section, key, int(value))

    def get_section(self, section: str, is_dict: bool = True) -> Union[list[tuple[str, str]], dict]:
        output = self.config.items(section)
        if is_dict:
            pre_output = dict()
            for el in output:
                pre_output[el[0]] = el[1]
            output = pre_output
        return output

    def edit(self, block: dict[str, dict[str, Any]], rewrite: bool = False) -> None:
        if rewrite:
            config = ConfigParser()
            for section, options in block.items():
                config[section] = options
            with open('output.ini', 'w') as configfile:
                config.write(configfile)
        else:
            for section, options in block.items():
                self.edit_section(section, options)

    def edit_section_if_not_none(self, section: str, block: dict, rewrite: bool = False) -> None:
        output = dict()
        for key, value in block.items():
            if value not in (None, 'None'):
                output[key] = value
        self.edit_section(section, output, rewrite)

    def edit_if_not_none(self, block: dict[str, dict[str, Any]], rewrite: bool = False) -> None:
        for section, options in block.items():
            self.edit_section_if_not_none(section, options, rewrite)

    def __getitem__(self, section):
        return self._Section(self.config, section)


class JavaScriptObjectNotationFile(TextFile):
    class JSONError(Exception):
        pass

    def __init__(self, path: str, encoding: Union[str, int] = 'utf-8'):
        super().__init__(path, encoding)
        extension_check(self.path, 'json')
        self.data = json.loads(self.read())

    def __dict__(self):
        return self.data

    def __len__(self):
        return len(self.data)

    def __bool__(self):
        return bool(self.data)

    def save(self):
        with open(self.path, 'w', encoding=self.encoding) as file:
            json.dump(self.data, file, indent=4, ensure_ascii=self.encoding.islower() == 'ascii')

    def get_value(self, key, default=KeyError()):
        try:
            if type(default) is KeyError:
                raise default
            return self.data.get(key, default)
        except KeyError:
            raise self.JSONError(f"В файле {get_filename(self.path)} не существует ключа {key}")

    def set_value(self, key, value):
        self.data[key] = value
        self.save()

    def delete_value(self, key):
        del self.data[key]
        self.save()

    def __setitem__(self, key, value):
        self.data[key] = value
        self.save()

    def __getitem__(self, key):
        return self.data[key]

    def __delitem__(self, key):
        del self.data[key]
        self.save()

    def __iter__(self):
        return iter(self.data)

    def set_values(self, **values):
        for key, value in values.items():
            self.data[key] = value
        self.save()

    def get_values(self, *keys):
        output = dict()
        for key in keys:
            output[key] = self.get_value(key)
        return output

    def key_exists(self, key) -> bool:
        return key in self.data

    def add_dict(self, data: dict):
        self.set_values(**data)

    def clear(self):
        self.data.clear()
        self.save()

    def rewrite_dict(self, data: dict):
        self.clear()
        self.add_dict(data)

    def rewrite_values(self, **values):
        self.rewrite_dict(values)

    def __str__(self):
        return str(self.data)

    def read_data(self):
        return self.data


def write_file_if_not_exists(path: str, text: str = ''):
    if os.path.isfile(path):
        raise FileExistsError(f'[WinError 183] Невозможно создать новый файл, так как он уже существует: {path}')
    else:
        create_file(path)
        file = TextFile(path)
        file.rewrite(text)


class CommaSeparatedValuesFile(TextFile):
    def __init__(self, path: str, encoding: Union[str, int] = 'utf-8'):
        super().__init__(path, encoding)

    def read_data(self):
        with self.open() as file:
            return csv.DictReader(file)

    def write_data(self, data: List[dict[str, Any]]):
        with self.open('w', newline='') as file:
            fields = equals_keys(*data)
            if fields is None:
                raise ValueError
            writer = csv.DictWriter(file, fieldnames=fields)
            writer.writeheader()
            writer.writerows(data)

# class CascadeStyleSheetsFile(TextFile):
#     def __init__(self, path: str, encoding: Union[str, int] = 'utf-8'):
#         super().__init__(path, encoding)
#
#     def save(self):
#



import math
import time
from typing import Union, Optional
from hrenpack.listwork import intlist
from random import randint


class HexagonalError(ValueError):
    pass


def int_float_separate(input: float, is_tuple: bool = False, is_int: bool = False):
    int_and_float = str(input).split('.')
    if is_int:
        int_and_float = intlist(int_and_float)
    if is_tuple:
        int_and_float = tuple(int_and_float)
    return int_and_float


def is_square_int(integer: int):
    sq = math.sqrt(integer)
    ls = int_float_separate(sq, True)
    return ls[1] == '0'


def division_with_rounding(dividend: float, divisor: float, round_in_any_case: bool = False, round_according_to_the_laws_of_mathematics: bool = False):
    quotient = dividend / divisor
    if round_in_any_case:
        output = int_float_separate(quotient, True, True)[0]
    elif round_according_to_the_laws_of_mathematics:
        output = round(quotient, 0)
    else:
        output = dividend // divisor
    return output


presence_of_remainder_on_division = lambda dividend, divisor: dividend % divisor == 0


# change - шанс
# Шанс указывается в процентах
def true_chance(chance: float):
    b = int(1 / (chance / 100))
    rand = randint(1, b)
    return rand == b


def number_in(num: float, minimum: float, maximum: float, steel: bool = False) -> bool:
    if minimum >= maximum:
        raise ValueError(f'Минимум должен быть меньше максимума')
    else:
        return minimum < num < maximum if steel else minimum <= num <= maximum


def hex_to_dec(hex_string: str) -> int:
    return int(hex_string, 16)


def dec_to_hex(dec: int) -> str:
    num = hex(dec)
    return num[2:]


def oct_to_dec(oct_int: int) -> int:
    return int(str(oct_int), 8)


def dec_to_oct(dec: int) -> int:
    num = oct(dec)
    return int(num[2:])


class Number:
    def __init__(self, dec: float):
        self.dec = dec
        self.hex = self.__hex__()
        self.oct = self.__oct__()

    def __hex__(self) -> str:
        if not '.' in str(self.dec):
            return dec_to_hex(int(self.dec))
        else:
            ot, to = str(self.dec).split('.')
            return f'{dec_to_hex(int(ot))}.{dec_to_hex(int(to))}'

    def __oct__(self) -> Union[int, float]:
        if not '.' in str(self.dec):
            return dec_to_oct(int(self.dec))
        else:
            ot, to = str(self.dec).split('.')
            return float(f'{dec_to_oct(int(ot))}.{dec_to_oct(int(to))}')

    def __int__(self) -> Union[int, float]:
        return self.dec

    def __str__(self) -> str:
        return str(self.dec)

    def __add__(self, other):
        if type(other) is Number:
            return Number(self.dec + other.dec)
        elif type(other) is int or type(other) is float:
            return Number(self.dec + other)
        else:
            raise TypeError("Можно складывать только числа")

    def __sub__(self, other):
        if type(other) is Number:
            return Number(self.dec - other.dec)
        elif type(other) is int or type(other) is float:
            return Number(self.dec - other)
        else:
            raise TypeError("Можно вычитать только числа")

    def __mul__(self, other):
        if type(other) is Number:
            return Number(self.dec * other.dec)
        elif type(other) is int or type(other) is float:
            return Number(self.dec * other)
        else:
            raise TypeError("Можно умножать только числа")

    def __truediv__(self, other):
        if type(other) is Number:
            return Number(self.dec / other.dec)
        elif type(other) is int or type(other) is float:
            return Number(self.dec / other)
        else:
            raise TypeError("Можно делить только числа")

    def __floordiv__(self, other):
        if type(other) is Number:
            return Number(self.dec // other.dec)
        elif type(other) is int or type(other) is float:
            return Number(self.dec // other)
        else:
            raise TypeError("Можно делить только числа")

    def __mod__(self, other):
        if type(other) is Number:
            return Number(self.dec % other.dec)
        elif type(other) is int or type(other) is float:
            return Number(self.dec % other)
        else:
            raise TypeError("Можно делить только числа")

    def __divmod__(self, other):
        if type(other) is Number:
            return Number(self.dec % other.dec)
        elif type(other) is int or type(other) is float:
            return Number(self.dec % other)
        else:
            raise TypeError("Можно делить только числа")

    def __pow__(self, power, modulo=None):
        if type(power) is Number:
            return Number(self.dec ** power.dec)
        elif type(power) is int or type(power) is float:
            return Number(self.dec ** power)
        else:
            raise TypeError("Можно делить только числа")

    def __lt__(self, other) -> bool:
        if type(other) is Number:
            return self.dec < other.dec
        elif type(other) is int or type(other) is float:
            return self.dec < other
        else:
            raise TypeError("Можно сравнивать только числа")

    def __le__(self, other) -> bool:
        if type(other) is Number:
            return self.dec <= other.dec
        elif type(other) is int or type(other) is float:
            return self.dec <= other
        else:
            raise TypeError("Можно сравнивать только числа")

    def __gt__(self, other) -> bool:
        if type(other) is Number:
            return self.dec > other.dec
        elif type(other) is int or type(other) is float:
            return self.dec > other
        else:
            raise TypeError("Можно сравнивать только числа")

    def __ge__(self, other) -> bool:
        if type(other) is Number:
            return self.dec >= other.dec
        elif type(other) is int or type(other) is float:
            return self.dec >= other
        else:
            raise TypeError("Можно сравнивать только числа")

    def __eq__(self, other) -> bool:
        if type(other) is Number:
            return self.dec == other.dec
        elif type(other) is int or type(other) is float:
            return self.dec == other
        else:
            raise TypeError("Можно сравнивать только числа")

    def __ne__(self, other) -> bool:
        if type(other) is Number:
            return self.dec != other.dec
        elif type(other) is int or type(other) is float:
            return self.dec != other
        else:
            raise TypeError("Можно сравнивать только числа")

    def __bool__(self) -> bool:
        return bool(self.dec)


fln = Union[float, Number]


def moreless(num: fln, min: fln, max: fln, is_strict: bool = False) -> bool:
    if min > max:
        raise ValueError("min должен быть меньше max")
    return min < num < max if is_strict else min <= num <= max


moreless_strict = lambda num, min, max: moreless(num, min, max, True)
moreless_not_strict = lambda num, min, max: moreless(num, min, max, False)


def pifs(number: Union[fln, str]) -> str:
    if type(number) is str:
        number = float(number)
    return str(number) if number <= 0 else f'+{number}'


def closest_number(number, *numbers, prefer_max: Optional[bool] = None):
    closest_distance = min(abs(x - number) for x in numbers)
    candidates = [x for x in numbers if abs(x - number) == closest_distance]
    if prefer_max is None:
        return candidates[0]
    elif prefer_max:
        return max(candidates)
    else:
        return min(candidates)


def to_fahrenheit(temp_celsius: float, round_: int = 0):
    return round((temp_celsius * 9 / 5) + 32, round_)


def to_celsius(temp_fahrenheit: float, round_: int = 0):
    return round((temp_fahrenheit - 32) * 5 / 9, round_)


def zero_len(number: float, length: int) -> str:
    number = str(number)
    integer, fl = number.split('.')
    integer = '0' * (length - len(integer)) + integer
    return '.'.join([integer, fl])


def module(number: float):
    return number if number >= 0 else -number


def round_and_delete(number: int, digits: int):
    d = int('1' + '0' * digits)
    number = round(number, -digits)
    return number // d



import functools, inspect
from abc import ABC, abstractmethod, abstractproperty, abstractclassmethod, abstractstaticmethod

from hrenpack.functionwork import empty_function
from hrenpack.listwork import get_from_dict, _is_tuple


class EncapsulationError(Exception):
    pass


def count_inheritance_levels(cls):
    # Получаем все классы в порядке разрешения методов
    mro = cls.__mro__
    # Считаем, сколько раз base_class есть в цепочке MRO
    count = 0
    for c in mro:
        if c is object:
            count += 1
    return count - 2


def check_method_in_parent(cls, name, debug_mode: bool = False):
    bases = cls.__bases__
    if debug_mode:
        print(bases)
    for base in bases:
        if hasattr(base, name):
            if debug_mode:
                print(base.__name__)
            return True
    return False


class AbstractClass(ABC):
    def __new__(cls, *args, **kwargs):
        if cls is AbstractClass or AbstractClass in cls.__bases__:
            raise TypeError("Это абстрактный класс, его можно только наследовать")
        try:
            return super().__new__(cls, *args, **kwargs)
        except TypeError:
            return super().__new__(cls)


# def abstractclass(cls):
#     class _Class(cls, AbstractClass):
#         pass
#     return _Class


# class ClassMethodMeta(type):
#     """Не работает"""
#     def __new__(cls, name, bases, attrs):
#         new_class = super().__new__(cls, name, bases, attrs)
#
#         for attr_name, attr_value in attrs.items():
#             if callable(attr_value):
#                 if '__' in attr_name:
#                     continue
#                 setattr(new_class, attr_name, classmethod(attr_value))
#
#         return new_class
#
#
# class StaticClass(metaclass=ClassMethodMeta):
#     """Не работает"""
#     def __new__(cls):
#         raise TypeError("Это статический класс, он не может создавать объекты")
#
#
# def staticclass(cls):
#     """Не работает"""
#     class _Class(StaticClass, cls):
#         pass
#     return _Class


def inner_function():
    frame = inspect.currentframe()
    local_vars = frame.f_back.f_locals
    del frame
    return local_vars


def protectedmethod(method):
    @functools.wraps(method)
    def wrapper(*args, **kwargs):
        stack = inspect.stack()
        for frame in stack:
            caller = frame.frame
            if caller.f_locals.get('self') is args[0]:
                return method(*args, **kwargs)
        raise EncapsulationError("Это защищенный метод")
    return wrapper


def privatemethod(method):
    """Пока что работает только как защищенный метод"""
    return protectedmethod(method)


def supermethod(method):
    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        getattr(super(type(self), self), method.__name__, empty_function)(*args, **kwargs)
        return method(self, *args, **kwargs)
    return wrapper


def supermethod_post(method):
    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        output = method(self, *args, **kwargs)
        getattr(super(type(self), self), method.__name__, empty_function)(*args, **kwargs)
        return output
    return wrapper


def superonlymethod(method):
    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        return getattr(super(type(self), self), method.__name__, empty_function)(*args, **kwargs)
    return wrapper


def addattr(instance, attr_name, value):
    if not hasattr(instance, attr_name):
        setattr(instance, attr_name, value)


def update_attrs_from_dict(instance, **attrs):
    for attr_name, value in attrs.items():
        setattr(instance, attr_name, value)


def add_attrs_from_dict(instance, **attrs):
    for attr_name, value in attrs.items():
        addattr(instance, attr_name, value)


def setattr_if_is_none(instance, attr_name, value):
    if getattr(instance, attr_name, None) is None:
        addattr(instance, attr_name, value)


def set_attrs_if_is_none(instance, **attrs):
    for attr_name, value in attrs.items():
        setattr_if_is_none(instance, attr_name, value)


def getattrs(instance, *attr_names, only_values: bool = False, is_tuple: bool = False, default=None):
    output = dict()
    for attr_name in attr_names:
        output[attr_name] = getattr(instance, attr_name, default)
    return get_from_dict(output, *output.keys(), is_tuple=is_tuple, only_values=only_values, default=default)


class SafeInheritance:
    """Базовый класс, который предотвращает ошибки MRO при наследовании."""

    def __new__(cls, name, bases, namespace):
        try:
            # Пробуем создать класс с текущими базовыми классами
            new_class = super().__new__(cls, name, bases, namespace)
            # Проверяем MRO (если есть конфликт, вызовется TypeError)
            new_class.mro()
            return new_class
        except TypeError:
            # Если MRO невалиден, наследуем только от object (или другой резервный класс)
            print(f"Warning: MRO conflict in {name}. Falling back to basic inheritance.")
            return super().__new__(cls, name, (object,), namespace)


class SafeMeta(type):
    """Метакласс для безопасного наследования."""

    def __new__(mcls, name, bases, namespace):
        try:
            new_class = super().__new__(mcls, name, bases, namespace)
            new_class.mro()  # Проверяем MRO
            return new_class
        except TypeError:
            print(f"Warning: MRO conflict in {name}. Using fallback inheritance.")
            return super().__new__(mcls, name, (object,), namespace)



from hrenpack.classes import DataClass

file_dialog_templates = DataClass(images="Изображения (*.jpg *.jpeg *.png *.tif *.tiff)", all="Все файлы (*)",
                                  txt="Текстовый документ (*.txt)", srt="Субтитры SRT (*.srt)")


import string, re
from typing import Union
from random import randint, choice as randchoice
from hrenpack.listwork import split_list, ab_reverse, tuplist, _is_tuple

PYTHONNAME_LETTERS = string.ascii_lowercase + string.digits + '_'


def tuple_to_str(*args, letter: str = ',') -> str:
    output = ''
    for arg in args:
        if output == '':
            output = arg
        else:
            output = f'{output}{letter} {arg}'
    return output


def antienter_str(string: str, space: bool = True) -> str:
    string_list = string.split('\n')
    output = ''
    if space:
        for s in string_list:
            output = f'{output} {s}'
    else:
        for s in string_list:
            output = output + s
    return output


def zap_list(string: str, letter: str = ',', is_tuple: bool = False) -> list:
    output = string.split(letter)
    return tuple(output) if is_tuple else output


def antispace(string: str) -> str:
    space_list = string.split()
    output = ''
    for el in space_list:
        output = output + el
    return output


def string_add(*args: str) -> str:
    output = ''
    for s in args:
        output = output + s
    return output


def if_empty_str(string: str, empty: str, not_empty: str) -> str:
    return empty if string == '' else not_empty


def randstr(a: int, b: int) -> str:
    return str(randint(a, b))


def search_and_edit(text: str, input: str, output: str) -> str:
    if input in text:
        frags = text.split(input)
        if frags.__len__() == 1:
            fragment = frags[0]
            if is_first(text, input):
                new_text = output + fragment
            elif is_last(text, input):
                new_text = fragment + output
            else:
                new_text = fragment
        else:
            new_text = split_list(frags, output)
            if is_first(text, input):
                new_text = output + new_text
            elif is_last(text, input):
                new_text = new_text + output
        return new_text
    else:
        return text


enter_fix = lambda text: search_and_edit(text, '\\n', '\n')


def prefix(base: str, prefix: str, is_suffix: bool = False) -> str:
    return '{} {}'.format(*ab_reverse(base, prefix, is_suffix, True))


def in_or(string: str, *args: str) -> bool:
    output = list()
    for arg in args:
        output.append(arg in string)
    return any(output)


def in_and(string: str, *args: str) -> bool:
    output = list()
    for arg in args:
        output.append(arg in string)
    return all(output)


def index_edit(string: str, index: int, letter: str) -> str:
    if len(letter) != 1:
        raise ValueError('Letter must be a single character')
    p1 = string[:index]
    p2 = string[index + 1:]
    return p1 + letter + p2


def string_reverse(string: str):
    output = list(string)
    output.reverse()
    return ''.join(output)


def index_edit_multi(string: str, values: dict[int, str]) -> str:
    for key, value in values.items():
        string = index_edit(string, key, value)
    return string


def index_edit_join(string: str, indexes: tuplist, values: Union[tuplist, str]) -> str:
    if len(indexes) != len(values):
        raise ValueError('Length of indexes must be equal to length of values')
    elif len(indexes) == 0 or len(values) == 0:
        raise ValueError('Empty indexes or values')
    else:
        for i in range(len(indexes)):
            if type(indexes[i]) is not int:
                raise ValueError('Indexes must be integers')
            if type(values[i]) is not str:
                raise ValueError('Values must be strings')
            string = index_edit(string, indexes[i], values[i])
        return string


def generate_rand_string(length: int = 25):
    return ''.join(randchoice(string.ascii_letters + string.digits) for _ in range(length))


def unspace(input: str):
    return input.replace(' ', '')


def unspace_multi(*strs: str) -> tuple:
    return tuple(unspace(s) for s in strs)


def remove_extra_spaces(text):
    # Заменяем несколько пробелов подряд на один
    return re.sub(r'\s+', ' ', text).strip()


def words_to_letters(*words: str, is_tuple: bool) -> tuplist:
    return _is_tuple(tuple(''.join(words)), is_tuple)


def only_this_letters(text: str, *letters: str) -> bool:
    letters = words_to_letters(*letters)
    for letter in text:
        if letter not in letters:
            return False
    return True


def only_pythonname(text: str) -> bool:
    return only_this_letters(text, PYTHONNAME_LETTERS)



import platform
from logging import debug
from typing import Optional
from kivy.app import App as KivyApp
from kivy.core.window import Window


class App(KivyApp):
    width: int = 800
    height: int = 800
    title: str = "Window"
    pos_x: int = 0
    pos_y: int = 0
    resizable: bool = True
    fullscreen: bool = False
    borderless: bool = False
    min_width: int = 0
    min_height: int = 0
    max_width: int = 32767
    max_height: int = 32767
    alpha: float = 0
    icon: Optional[str] = None

    def on_start(self):
        if self.alpha < 0 or self.alpha > 1:
            raise ValueError('Alpha must be between 0 and 1')
        Window.title = self.title
        Window.resizable = self.resizable
        Window.fullscreen = self.fullscreen
        Window.borderless = self.borderless
        Window.min_width = self.min_width
        Window.min_height = self.min_height
        Window.max_width = self.max_width
        Window.max_height = self.max_height
        Window.alpha = self.alpha
        if platform.system() == 'Windows':
            Window.size = (self.width, self.height)
            Window.pos = (self.pos_x, self.pos_y)
        if self.icon:
            Window.icon = self.icon
        debug("WindowMixin initialized")


def pos_hint(x: Optional[float] = None, y: Optional[float] = None,
             center_x: Optional[float] = None, center_y: Optional[float] = None):
    output = {'x': x, 'y': y, 'center_x': center_x, 'center_y': center_y}
    for key in ('x', 'y', 'center_x', 'center_y'):
        if output[key] is None:
            del output[key]
        elif output[key] > 1:
            output[key] /= 100
    return output


def rgba(red: int, green: int, blue: int, alpha: int = 100):
    if red > 255:
        raise ValueError("Red value must be less than 255")
    elif red < 0:
        raise ValueError("Red value must be greater than 0")
    elif green > 255:
        raise ValueError("Green value must be less than 255")
    elif green < 0:
        raise ValueError("Green value must be greater than 0")
    elif blue > 255:
        raise ValueError("Blue value must be less than 255")
    elif blue < 0:
        raise ValueError("Blue value must be greater than 0")
    elif alpha > 100:
        raise ValueError("Alpha value must be less than 100")
    elif alpha < 0:
        raise ValueError("Alpha value must be greater than 0")
    else:
        return red / 255, green / 255, blue / 255, alpha / 100


def remove_parents(*widgets):
    for widget in widgets:
        widget.remove_parent()



import platform
from logging import debug
from typing import Optional
from kivy.app import App
from kivy.core.image import Image as CoreImage
from kivy.core.window import Window
from kivy.graphics import Rectangle
from hrenpack.listwork import tuplist, key_in_dict
from hrenpack.filework.programming_language_work import KivyStylesFile


class WindowMixin:
    """При наследовании обязательно ставить перед классом App, иначе не сработает
    При использовании виджетов с динамическим размером, до изменения размеров окна размер виджета очень крупный"""
    width: int = 800
    height: int = 800
    title: str = "Window"
    pos_x: int = 0
    pos_y: int = 0
    resizable: bool = True
    fullscreen: bool = False
    borderless: bool = False
    min_width: int = 0
    min_height: int = 0
    max_width: int = 32767
    max_height: int = 32767
    alpha: float = 0
    icon: Optional[str] = None

    def on_start(self):
        if self.alpha < 0 or self.alpha > 1:
            raise ValueError('Alpha must be between 0 and 1')
        Window.title = self.title
        Window.resizable = self.resizable
        Window.fullscreen = self.fullscreen
        Window.borderless = self.borderless
        Window.min_width = self.min_width
        Window.min_height = self.min_height
        Window.max_width = self.max_width
        Window.max_height = self.max_height
        Window.alpha = self.alpha
        if platform.system() == 'Windows':
            Window.size = (self.width, self.height)
            Window.pos = (self.pos_x, self.pos_y)
        if self.icon:
            Window.icon = self.icon
        debug("WindowMixin initialized")


class WidgetMixin:
    def __init__(self, **kwargs):
        self._style: Optional[KivyStylesFile] = None
        if 'bind' in kwargs:
            self.bind(**kwargs['bind'])
            del kwargs['bind']
        if 'kvs_file' in kwargs:
            file = kwargs['kvs_file']
            if type(file) is KivyStylesFile:
                file.apply(self)
            elif type(file) is str:
                file = KivyStylesFile(file)
                file.apply(self)
            else:
                raise TypeError('kvs_file must be a KivyStylesFile object')

    def disable(self):
        self.disabled = True

    def enable(self):
        self.disabled = False

    @staticmethod
    def get_app():
        return App.get_running_app()


class LayoutMixin(WidgetMixin):
    def add_widgets(self, widgets: tuplist, *args, **kwargs):
        for widget in widgets:
            self.add_widget(widget, *args, **kwargs)

    def add(self, *widgets):
        self.add_widgets(widgets)

    def all_widgets(self):
        return self.children


class ImageBackgroundMixin:
    def __init__(self, **kwargs):
        with self.canvas.before:
            if key_in_dict(kwargs, 'bg_image'):
                self._bg_image = self._val(kwargs['bg_image'])
                del kwargs['bg_image']
            self._rect = Rectangle(size=self.size, pos=self.pos, texture=self._bg_image)
            self.bind(size=self._update_rect, pos=self._update_rect)

    def _update_rect(self, instance, value):
        self._rect.pos = instance.pos
        self._rect.size = instance.size

    @property
    def bg_image(self):
        return self._bg_image

    @bg_image.setter
    def bg_image(self, value):
        self._bg_image = self._val(value)
        self._rect.texture = self._bg_image

    @staticmethod
    def _val(value):
        if isinstance(value, CoreImage):
            bg_image = value.texture
        elif isinstance(value, str):
            bg_image = CoreImage(value).texture
        else:
            bg_image = value
        return bg_image



from markupsafe import Markup


class DjangoStyleFormMixin:
    def as_p(self, include_hidden_fields: bool = True):
        """Рендерит форму как последовательность тегов <p>"""
        html = []
        for field in self:
            if not include_hidden_fields and field.type in ['HiddenField', 'CSRFTokenField']:
                continue

            if field.type in ['HiddenField', 'CSRFTokenField']:
                html.append(str(field))
            else:
                # Основной field
                field_html = [str(field.label), f'\t{field}']

                if field.type in ['BooleanField']:
                    field_html.reverse()

                # Help text
                if hasattr(field, 'help_text') and field.help_text:
                    field_html.append(f'<span class="helptext">{field.help_text}</span>')

                # Errors
                if field.errors:
                    error_html = ['<ul class="errorlist">']
                    for error in field.errors:
                        error_html.append(f'<li>{error}</li>')
                    error_html.append('</ul>')
                    field_html.append(''.join(error_html))

                html.append(f'<p>\n\t{"\n".join(field_html)}\n</p>')

        return Markup('\n'.join(html))


import importlib, pkgutil
from importlib import import_module


def import_all_submodules(package_name):
    """Динамически импортирует все подмодули в пакете."""
    package = import_module(package_name)

    for _, module_name, _ in pkgutil.iter_modules(package.__path__):
        full_module_name = f"{package_name}.{module_name}"
        try:
            importlib.import_module(full_module_name)
        except ImportError as e:
            print(f"Failed to import {full_module_name}: {e}")



import socket
from dataclasses import dataclass


# class NetworkInterfaces:
#     @dataclass
#     class NetworkInterface:
#         pass
#
#     def __init__(self):
#         pass


def get_ipv4_addresses(exclude_localhost: bool = False) -> list:
    """Получить все IPv4 адреса локального хоста"""
    ip_addresses = []

    # Получаем имя хоста
    hostname = socket.gethostname()

    try:
        # Получаем все IP адреса, связанные с хостом
        addresses = socket.getaddrinfo(hostname, None)

        for addr_info in addresses:
            # addr_info[0] - семейство адресов (AF_INET для IPv4)
            # addr_info[4][0] - IP адрес
            if addr_info[0] == socket.AF_INET:  # IPv4
                ip = addr_info[4][0]
                if ip not in ip_addresses and (ip != '127.0.0.1' or not exclude_localhost):
                    ip_addresses.append(ip)

    except socket.gaierror:
        pass

    # Если не нашли адресов, пробуем другой способ
    if not ip_addresses:
        try:
            # Создаем временный сокет для получения локального IP
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect(('8.8.8.8', 80))  # Подключаемся к внешнему серверу
            local_ip = s.getsockname()[0]
            s.close()
            if local_ip != '127.0.0.1' or not exclude_localhost:
                ip_addresses.append(local_ip)
        except:
            pass

    return ip_addresses



import os, ctypes, shutil, getpass, platform, random, string, subprocess, warnings
from datetime import datetime
from hrenpack.listwork import split_list
from typing import Union, List
from dataclasses import dataclass
from pathlib import Path

PathType = Union[str, Path]


def get_filename(path: str, raise_error: bool = True) -> str:
    if '/' in path:
        output = path.split('/')[-1]
    elif '\\' in path:
        output = path.split('\\')[-1]
    else:
        output = path

    if not os.path.isfile(path) and raise_error:
        raise FileNotFoundError('No such file: ' + path)
    else:
        return output


def get_extension(path: str, raise_error: bool = True) -> str:
    filename = get_filename(path, raise_error)
    return filename.split('.')[-1] if '.' in filename else ''


def get_path_without_filename(path: str, raise_error: bool = True):
    v85 = True
    if '/' in path:
        path_list = path.split('/')
        path_list.pop()
    elif '\\' in path:
        path_list = path.split('\\')
        path_list.pop()
    else:
        v85 = False
    output = split_list(path_list, '/') if v85 else ''
    if not os.path.isfile(path) and raise_error:
        raise FileNotFoundError('No such file: ' + path)
    else:
        return output


def get_path_and_filename(path: str, raise_error: bool = True):
    if raise_error and not os.path.isfile(path):
        raise FileNotFoundError('No such file: ' + path)
    return get_path_without_filename(path), get_filename(path)


def rename(path: str, new_filename: str):
    pwfn, filename = get_path_and_filename(path)
    if pwfn:
        new_path = f'{pwfn}/{new_filename}'
    else:
        new_path = new_filename
    os.rename(path, new_path)


def create_file(path: str):
    try:
        open(path, 'x').close()
    except FileExistsError:
        raise FileExistsError(f'[WinError 183] Невозможно создать новый файл, так как он уже существует: {path}')


def get_filename_without_extension(path: str, raise_error: bool = True) -> str:
    filename = get_filename(path, raise_error)
    fl = filename.split('.')
    fl.pop()
    return split_list(fl, '.')


def get_path_without_extension(path: str, raise_error: bool = True) -> str:
    filename = get_filename_without_extension(path, raise_error)
    pwfl = get_path_without_filename(path, raise_error)
    return f'{pwfl}/{filename}'


@dataclass
class FileNameInfo:
    path: str
    filename: str = ''
    extension: str = ''
    path_without_extension: str = ''
    path_without_filename: str = ''

    def __post_init__(self):
        self.filename = get_filename(self.path)
        self.extension = get_extension(self.path)
        self.path_without_extension = get_path_without_filename(self.path)
        self.path_without_filename = get_path_without_extension(self.path)


def delete_file(path: str):
    os.remove(path)


def create_file_exist(path: str, space: bool = True, return_filename_and_path: bool = False):
    if not os.path.isfile(path):
        new_path = path
    else:
        counter = 0
        separator = ' ' if space else ''
        pafn = get_path_without_extension(path, False)
        extension = get_extension(path)
        while True:
            new_path = f'{pafn}{separator}({counter}).{extension}'
            if os.path.isfile(new_path):
                counter += 1
            else:
                break

    create_file(new_path)

    if return_filename_and_path:
        return FileNameInfo(new_path)
    return None


def edit_time(year: int = -1, month: int = -1, day: int = -1, hour: int = -1, minute: int = -1,
              second: int = -1) -> None:
    now = datetime.now()
    if year < 0:
        year = now.year
    if month < 0:
        month = now.month
    if day < 0:
        day = now.day
    if hour < 0:
        hour = now.hour
    if minute < 0:
        minute = now.minute
    if second < 0:
        second = now.second

    os.system(f'date {year}.{month}.{day}')
    os.system(f'time {hour}:{minute}:{second}')


def is_admin() -> bool:
    return bool(ctypes.windll.shell32.IsUserAnAdmin())


def admin_error() -> None:
    if not is_admin():
        raise OSError("Перезапустите программу с правами администратора")


def admin_pause() -> None:
    if not is_admin():
        print("Перезапустите программу с правами администратора")
        input()


def admin_pause_exit() -> None:
    admin_pause()
    exit(1)


def remove_files_and_folders(*paths):
    for path in paths:
        if os.path.isfile(path):
            try:
                os.remove(path)
            except Exception as e:
                raise OSError(f"Не удалось удалить файл {path}: {e}")
        elif os.path.isdir(path):
            try:
                shutil.rmtree(path)
            except Exception as e:
                raise OSError(f"Не удалось удалить папку {path}: {e}")


def get_username() -> str:
    return getpass.getuser()


def android_path(path: str, domain: str, name: str) -> str:
    if platform.system() == 'Windows':
        return path
    return f'data/data/{domain}.{name}/files/app/{path}'


class AndroidPath:
    def __init__(self, domain: str, name: str):
        self.domain = domain
        self.name = name

    def __call__(self, path: str) -> str:
        if platform.system() == 'Windows':
            return path
        return f'data/data/{self.domain}.{self.name}/files/app/{path}'


def get_files_startswith(directory: str, start: str, full_path: bool = True) -> List[str]:
    files = os.listdir(directory)
    filtered_files = [f for f in files if f.startswith(start)]
    if full_path:
        output = list()
        for file in filtered_files:
            output.append(os.path.join(directory, file))
    else:
        output = filtered_files
    return output


def all_files_and_dirs(directory):
    output = []
    for root, dirs, files in os.walk(directory):
        output.extend([os.path.join(root, name) for name in files])
        output.extend([os.path.join(root, name) for name in dirs])
    return output


def all_files(directory, full_path: bool = True) -> List[str]:
    path = Path(directory)
    if not full_path:
        return [get_filename(str(file)) for file in path.rglob('*') if file.is_file()]
    return [str(file) for file in path.rglob('*') if file.is_file()]


def generate_random_filename(length: int = 10, extension: str = '') -> str:
    letters_and_digits = string.ascii_letters + string.digits
    random_name = ''.join(random.choice(letters_and_digits) for _ in range(length))
    if extension:
        extension = '.' + extension
    return random_name + extension


def compare_versions(version1, version2):
    version1_parts = list(map(int, version1.split('.')))
    version2_parts = list(map(int, version2.split('.')))

    # Сравниваем по частям
    for v1, v2 in zip(version1_parts, version2_parts):
        if v1 > v2:
            return 1
        elif v1 < v2:
            return -1
    return 0


def uninstall_program(program_name):
    try:
        # Выполняем команду для деинсталляции программы
        subprocess.run(['wmic', 'product', 'where', f'name="{program_name}"', 'call', 'uninstall'], check=True)
        print(f'Программа "{program_name}" успешно деинсталлирована.')
    except subprocess.CalledProcessError as e:
        print(f'Ошибка при деинсталляции программы: {e}')


def admin_required(func):
    def wrapper(*args, **kwargs):
        if is_admin():
            return func(*args, **kwargs)
        raise OSError("Отказано в доступе")
    return wrapper


def package_is_debug(file: Path, tree_level: int = 1):
    if not file.is_file() or not file.exists() or get_extension(str(file)) != 'py':
        raise FileNotFoundError
    for level in range(tree_level + 1):
        file = file.parent
    if not file.is_dir():
        raise NotADirectoryError
    return file.name != 'site-packages'


class PackageIsDebug:
    def __init__(self, file: Union[Path, str], tree_level: int = 1):
        file = file if isinstance(file, Path) else Path(file)
        if not file.is_file() or not file.exists() or get_extension(str(file)) != 'py':
            raise FileNotFoundError
        self.file = file
        self.tree_level = tree_level

    def get_directory(self):
        file = self.file
        for level in range(self.tree_level + 1):
            file = file.parent
        if not file.is_dir():
            raise NotADirectoryError
        return file

    def is_debug(self):
        return self.get_directory().name != 'site-packages'

    def chdir(self):
        os.chdir(self.get_directory())



from wtforms import SelectField
from wtforms.validators import ValidationError


class BooleanSelectField(SelectField):
    """
    SelectField, который принимает 'true'/'false' в HTML,
    но возвращает bool в data
    """

    def __init__(self, label='', validators=None,
                 false_label: str = "Ложь",
                 true_label: str = "Истина",
                 **kwargs):
        # Устанавливаем стандартные choices
        choices = [('false', false_label), ('true', true_label)]

        # Позволяем переопределить choices, если нужно
        if 'choices' not in kwargs:
            kwargs['choices'] = choices

        # Не используем coerce=bool, так как он не работает с 'true'/'false'
        kwargs.pop('coerce', None)

        super().__init__(label, validators, **kwargs)

        # Сохраняем текстовые значения для валидации
        self.true_value = 'true'
        self.false_value = 'false'

    def process_formdata(self, valuelist):
        """
        Обрабатывает данные из формы
        """
        if valuelist:
            value = str(valuelist[0])
            # Преобразуем строку в bool
            if value == self.true_value:
                self.data = True
            elif value == self.false_value:
                self.data = False
            else:
                self.data = None
                raise ValidationError(f'Недопустимое значение: {value}')

            # ВАЖНО: сохраняем raw_data для валидации choices
            self.raw_data = [value]
        else:
            self.data = None
            self.raw_data = []

    def pre_validate(self, form):
        """
        Переопределяем валидацию, чтобы пропустить стандартную проверку choices
        Вместо этого проверяем, что значение True или False
        """
        # Пропускаем стандартную валидацию choices
        if self.data is not None and self.data in (True, False):
            return
        elif self.data is None and not self.raw_data:
            # Пустое значение
            pass
        else:
            raise ValidationError(self.gettext('Not a valid choice'))

    def _value(self):
        """
        Возвращает значение для отображения в HTML
        """
        if self.data is True:
            return self.true_value
        elif self.data is False:
            return self.false_value
        return ''



from typing import Any

from django.conf import settings
from django.contrib.auth.forms import UserCreationForm
from django.contrib.staticfiles.storage import staticfiles_storage
from django.db.models import Model
from django.shortcuts import render, redirect
from django.utils.decorators import classonlymethod
from django.views import View as DjangoView, generic
from django.contrib.auth import views as auth_views, logout, get_user_model, login, update_session_auth_hash
from django.contrib.auth.mixins import LoginRequiredMixin
from hrenpack import NullStr
from hrenpack.listwork import get_from_dict
from hrenpack.encapsulation import SafeInheritance, addattr, set_attrs_if_is_none
from hrenpack.framework.django import view_dict
from hrenpack.framework.django.forms import PasswordChangeForm, auth_forms
from hrenpack.framework.django.mixins import (NonAbstractMixin, ModelManagerMixin, UserAuthorizeMixin,
                                              TemplateViewMixin)


class BaseView(DjangoView, NonAbstractMixin):
    title: NullStr = None
    h1_title: NullStr = None
    dont_header: bool = False
    _base_template_name: str = getattr(settings, 'BASE_TEMPLATE', 'empty.html')

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        kwargs['title'] = self.title if self.title else "Страница"
        kwargs['h1_title'] = self.h1_title
        kwargs['dont_header'] = self.dont_header
        kwargs['base_template_name'] = self.base_template_name
        return view_dict(**kwargs)

    @property
    def base_template_name(self):
        return getattr(self, '_base_template_name',
                       getattr(settings, 'BASE_TEMPLATE', 'empty.html'))

    @base_template_name.setter
    def base_template_name(self, value):
        self._base_template_name = value or getattr(settings, 'BASE_TEMPLATE', 'empty.html')

    @classonlymethod
    def as_view(cls, **initkwargs):
        initkwargs.setdefault('title', "Страница")
        set_attrs_if_is_none(cls, **get_from_dict(initkwargs, 'title', 'h1_title', 'template_name',
                                                  'extra_context', pop_mode=True))
        return super().as_view(**initkwargs)


class View(BaseView, TemplateViewMixin):
    pass


class TemplateView(BaseView, generic.TemplateView):
    pass


class ListView(BaseView):
    model: Model
    context_name: str = 'db'

    def get_context_data(self, **kwargs):
        kwargs = super().get_context_data(**kwargs)
        kwargs[self.context_name] = self.get_queryset()
        return kwargs

    def get_queryset(self):
        return self.model.objects.all()

    # def get_model_dk(self, field: str):
    #     """Возвращает элементы из определенного поля модели"""
    #     if field in self.model._meta.fields:
    #         output = list()
    #         for el in self.get_queryset():
    #             output.append(el.__dict__[field])
    #         return output
    #     else:
    #         raise AttributeError("Указанного поля модели не существует")
    #
    # def get_object_or_404(self, field: str, value) -> Model:
    #     """Возвращает объект модели. Если нет, то возвращает исключение 404"""
    #     elements = self.get_queryset()
    #     values = self.get_model_dk(field)
    #     if not value in values:
    #         raise Http404
    #     return elements.get(**{field: value})


class DetailView(ModelManagerMixin, BaseView, generic.DetailView):
    pass


class FormView(BaseView, generic.FormView):
    pass


class CreateView(BaseView, generic.CreateView):
    pass


class UpdateView(ModelManagerMixin, BaseView, generic.UpdateView):
    pass


class PasswordChangeView(BaseView, auth_views.PasswordChangeView):
    pass


class PasswordChangeDoneView(BaseView, auth_views.PasswordChangeDoneView):
    pass


class PasswordResetView(BaseView, auth_views.PasswordResetView):
    pass


class PasswordResetDoneView(BaseView, auth_views.PasswordResetDoneView):
    pass


class PasswordResetConfirmView(BaseView, auth_views.PasswordResetConfirmView):
    pass


class PasswordResetCompleteView(UserAuthorizeMixin, BaseView, auth_views.PasswordResetCompleteView):
    pass


class LoginView(BaseView, auth_views.LoginView):
    title = "Авторизация"


class LogoutView(BaseView, auth_views.LogoutView):
    title = "Вы вышли из аккаунта"


def create_logout_view(template_name: str, title: str = "Вы вышли из аккаунта", h1_title: NullStr = None,
                       dont_header: bool = False, **kwargs):
    def logout_view(request):
        logout(request)
        return render(request, template_name, view_dict(title, h1_title, dont_header=dont_header, **kwargs))
    return logout_view


def create_logout_view_with_next():
    def logout_view(request):
        if request.user.is_authenticated:
            logout(request)
        return redirect(request.GET.get('next', '/'))
    return logout_view


class RegistrationView(CreateView):
    title = "Регистрация"
    model = get_user_model()
    form_class = UserCreationForm

    def form_valid(self, form):
        response = super().form_valid(form)
        login(self.request, self.object)
        return response


# class EditProfileView(View, LoginRequiredMixin):
#     profile_form_class: Any
#     password_change_form_class: Any = PasswordChangeForm
#     success_url: Any
#     title = "Изменить настройки пользователя"
#
#     def __new__(cls, *args, **kwargs):
#         if not issubclass(cls.password_change_form_class, PasswordChangeForm):
#             raise TypeError
#         return super().__new__(cls, *args, **kwargs)
#
#     def get(self, request, *args, **kwargs):
#         user_profile_form = self.profile_form_class(instance=request.user)
#         password_change_form = self.password_change_form_class(user=request.user)
#         return render(request, self.template_name, view_dict(
#             self.title, self.h1_title,
#             user_profile_form=user_profile_form,
#             password_change_form=password_change_form,
#             pur=True
#         ))
#
#     def post(self, request, *args, **kwargs):
#         user_profile_form = self.profile_form_class(request.POST, instance=request.user)
#         password_change_form = self.password_change_form_class(user=request.user, data=request.POST)
#         password_form_is_empty = self._password_form_is_empty(password_change_form)
#
#         if user_profile_form.is_valid() and any((password_change_form.is_valid(), password_form_is_empty)):
#             user_profile_form.save()
#             if password_form_is_empty:
#                 user = password_change_form.save()
#                 update_session_auth_hash(request, user)  # Обновляем сессию, чтобы не разлогинивать пользователя
#             return redirect(self.success_url)  # Замените на нужный вам URL
#
#         return render(request, self.template_name, view_dict(
#             self.title, self.h1_title,
#             user_profile_form=user_profile_form,
#             password_change_form=password_change_form,
#             pur=password_form_is_empty
#         ))
#
#     @staticmethod
#     def _password_form_is_empty(form):
#         cd = form.cleaned_data
#         return not cd['new_password1'] and cd['new_password2']
#
#
# class DirectoryView(View, NonAbstractMixin):
#     path: str
#
#     def get_context_data(self, **kwargs):
#         pass


class StaticFileView(generic.RedirectView):
    @classonlymethod
    def as_view(cls, **initkwargs):
        path = initkwargs.pop('path')
        if path is None:
            raise KeyError('path')
        return super().as_view(url=staticfiles_storage.url(path), **initkwargs)



import clipboard


class ClipBoardError(Exception):
    pass


def get_clipboard_image():
    from PIL import ImageGrab

    im = ImageGrab.grabclipboard()
    if im is not None:
        return im
    else:
        raise ClipBoardError("Буфер обмена пуст или не является изображением")


def copy_text(text: str) -> None:
    clipboard.copy(text)


def insert_text() -> str:
    return clipboard.paste()


def clipboard_is_image() -> bool:
    from PIL import ImageGrab
    im = ImageGrab.grabclipboard()
    return im is not None


def clipboard_image_error() -> None:
    if not clipboard_is_image():
        raise ClipBoardError("Буфер обмена пуст или не является изображением")


if __name__ == '__main__':
    clipboard_image_error()



from typing import Iterable, Literal


def multi_in(input: Iterable, *args, condition: Literal['or', 'and'] = 'or'):
    if not args:
        raise ValueError('Must provide at least one argument')
    for arg in args:
        if arg not in input and condition == 'and':
            return False
        elif arg in input and condition == 'or':
            return True
    return True



import re
from typing import Union, Literal, Optional, Iterable

tuplist = Union[tuple, list]
tdl = Union[tuple, dict, list]


def _is_tuple(input: Iterable, is_tuple: bool) -> Union[tuple, list]:
    return tuple(input) if is_tuple else list(input)


def antizero(wnull):
    if wnull < 10:
        output = '0' + str(wnull)
    else:
        output = str(wnull)
    return output


def listsearch(list, search):
    for i in range(len(list)):
        if list[i] == search:
            index = i
            break
        else:
            index = False
    return index


def antienter(list):
    o = ''
    for i in range(len(list)):
        o = o + list[i]
    output = o.split('\n')
    return output


def antienter_plus(list):
    o = ''
    for i in range(len(list)):
        o = o + list[i]
    output = o.split('\n')
    el0 = output[0].split('\ufeff')
    pel = el0[-1]
    output[0] = pel
    output.pop()
    return output


def intlist(input: list) -> list:
    for i in range(len(input)):
        input[i] = int(input[i])
    return input


def floatlist(input: tuplist) -> list:
    input = list(input)
    for i in range(len(input)):
        input[i] = float(input[i])
    return input


def list_add(input: list, index: int, data) -> list:
    forcikl = len(input) + 1
    output = list()
    for i in range(forcikl):
        if i < index:
            output.append(input[i])
        elif i == index:
            output.append(data)
        elif i > index:
            m = i - 1
            output.append(input[m])
    return output


def str_to_list_one(string: str) -> list:
    output = list()
    for letter in string:
        output.append(letter)
    return output


def in_number_series(ot: int, to: int, num: int) -> bool:
    number_series = list()
    for i in range(ot, to):
        number_series.append(i)
    return num in number_series


def in_numbers(ot: int, to: int, nums: Union[tuple, dict, list]) -> dict:
    is_dict = dict()
    for num in nums:
        is_dict[num] = in_number_series(ot, to, num)
    return is_dict


def dict_to_list(input: dict, is_tuple: bool):
    output = list()
    for key in input:
        value = input[key]
        output.append(value)
    return _is_tuple(output, is_tuple)


def multi_pop(input: list, *indexes: int):
    for i in range(len(indexes)):
        index = indexes[i] - i
        try:
            if index == len(input) - 1:
                input.pop()
            else:
                input.pop(index)
        except IndexError:
            pass
    return input


def if_dict_key(dct: dict, key):
    try:
        dct[key]
    except KeyError:
        return False
    else:
        return True


def split_list(input: Union[tuple, list], separator: str = ', '):
    return separator.join(input)


def merging_dictionaries(*dicts: dict) -> dict:
    output = dict()
    for d in dicts:
        output = {**output, **d}
    return output


# -1 - pseudo dict to dict
# 0 - выполнять в любом случае
# 1 - dict to pseudo dict
def pseudo_dictionary(input: tdl, not_edit_type: Literal[-1, 0, 1] = 0) -> tdl:
    if type(input) is dict and not_edit_type in (0, 1):
        output = list()
        for key, value in input.items():
            output.append((key, value))
    elif type(input) in (list, tuple) and not_edit_type in (-1, 0):
        output = dict()
        for el in input:
            key, value = el
            output[key] = value
    else:
        output = input
    return output


def variable_in_tuple(variable, *values):
    return variable in values


def pop(input: list, index: int = -1):
    input.pop(index)
    return input


split_list_enter = lambda input: split_list(input, '\n')
split_list_space = lambda input: split_list(input, ' ')
split_list_tab = lambda input: split_list(input, '\t')
dict_key_null = lambda dct, key: dct[key] if key in dct else None
dict_key_false = lambda dct, key: dct[key] if key in dct else False
dict_key_tf = lambda dct, key, true, false: true if key in dct else false
dict_keyf = lambda dct, key, false: dct[key] if key in dct else false
key_in_dict = lambda dct, key: dict_key_tf(dct, key, True, False)


def ab_reverse(a, b, condition: bool, is_tuple: bool = True) -> tuplist:
    output = [a, b]
    if condition:
        output.reverse()
    return tuple(output) if is_tuple else output


def ab_not_reverse(a, b, condition: bool, is_tuple: bool = True) -> tuplist:
    output = ab_reverse(a, b, condition, is_tuple)
    output.reverse()
    return output


def multi_reverse(condition, *args, is_tuple: bool = True) -> tuplist:
    args = list(args)
    if condition:
        args.reverse()
    return tuple(args) if is_tuple else args


def multi_not_reverse(condition, *args, is_tuple: bool = True) -> tuplist:
    output = multi_reverse(condition, *args, is_tuple=is_tuple)
    output.reverse()
    return output


def dict_keys_values(keys: tuplist, values: tuplist) -> dict:
    return dict(zip(keys, values))


def dkv_dict(source: dict, keys: tuplist, values: tuplist) -> dict:
    return merging_dictionaries(source, dict_keys_values(keys, values))


def remove_all(input: list, value, is_tuple: bool = False):
    count = input.count(value)
    for i in range(count):
        input.remove(value)
    return _is_tuple(input, is_tuple)


def remove_multi(input: list, *values, _remove_all: bool = False, is_tuple: bool = False):
    if _remove_all:
        for value in values:
            input = remove_all(input, value)
    else:
        for value in values:
            input.remove(value)
    return _is_tuple(input, is_tuple)


def list_to_list(input: list, index: Optional[int] = None, element=None, is_tuple: bool = True) -> tuplist:
    if index is not None and element is not None:
        raise TypeError("Оба аргумента не равняются None")
    elif index is None:
        index = input.index(element)
    in2 = input[index:-1]
    in2.pop(0)
    in2.append(input[-1])
    return _is_tuple((input[0:index], in2), is_tuple)


def list_tuple_to_str(input) -> str:
    try:
        if input is None:
            output = '()'
        elif isinstance(input, list) or all((isinstance(input, tuple), len(input) > 1)):
            output = str(input)
        elif isinstance(input, tuple) and len(input) == 1:
            output = f'({input[0]})'
        else:
            output = f'({input})'
        return output
    except TypeError:
        return f'({input})'


def split_quotes(text: str, is_tuple: bool = False) -> tuplist:
    pattern = r"""
            (?:
                "(?:[^"\\]|\\.)*"
                |
                '(?:[^'\\]|\\.)*'
                |
                \S+              
            )
        """
    tokens = re.findall(pattern, text, re.VERBOSE)
    return _is_tuple(tokens, is_tuple)


def get_values_by_keys(input: dict, *keys, is_tuple: bool = False) -> tuplist:
    output = list()
    for key in keys:
        output.append(input[key])
    return _is_tuple(output, is_tuple)


def del_keys(input: dict, *keys) -> None:
    for key in keys:
        del input[key]


def dict_index(input: dict, value):
    for k, v in input.items():
        if v == value:
            return k
    else:
        raise ValueError(f"Словарь {input} не содержит значения {value}")


def strlist(input: tuplist, is_tuple: bool = False):
    input = list(input)
    for i in range(len(input)):
        input[i] = str(input[i])
    return _is_tuple(input, is_tuple)


def keys_dict_equals(*dicts: dict) -> bool:
    dicts = list(dicts)
    first = tuple(dicts.pop(0).keys())
    for d in dicts:
        if first != tuple(d.keys()):
            return False
    return True


def equals_keys(*dicts: dict) -> tuple:
    if keys_dict_equals(*dicts):
        return tuple(dicts[0].keys())


def del_none(*args, is_tuple: bool = False) -> tuplist:
    args = list(args)
    for i, arg in enumerate(args):
        if arg is None:
            args.pop(i)
    return _is_tuple(args, is_tuple)


def del_none_from_dict(*dicts, **kwargs) -> dict:
    kwargs = merging_dictionaries(*dicts, kwargs)
    output = kwargs.copy()
    for key, value in kwargs.items():
        if value is None:
            output.pop(key)
    return output


def enum_tuple(iterable):
    return enumerate(tuple(iterable))


def split_list_to_lists(input: tuplist, *indexes: int, is_tuple: bool = False,
                        in_start: bool = True, in_end: bool = False) -> list:
    input = list(input)
    output = list()
    for i in range(len(indexes)):
        cur = indexes[i]
        past = indexes[i - 1] if i > 0 else 0
        if not in_start and past > 0:
            past -= 1
        if in_end:
            cur += 1
        sl = input[:cur] if past == 0 else input[past:cur]
        output.append(sl)
    return _is_tuple(output, is_tuple)


def get_from_dict(input: dict, *keys, only_values: bool = False, is_tuple: bool = False, default=None,
                  pop_mode: bool = False) -> tdl:
    output = dict()
    for key in keys:
        value = input.pop(key, default)
        output[key] = value
        if not pop_mode:
            input[key] = value
    if only_values:
        return _is_tuple(output.values(), is_tuple)
    return output


def replace_fragment_from_args(old_frag: str, new_frag: str, *args: str, is_tuple: bool = False) -> tuplist:
    output = list()
    for arg in args:
        output.append(arg.replace(old_frag, new_frag))
    return _is_tuple(output, is_tuple)


class dict_enumerate:
    def __init__(self, items: dict):
        self.items = items.items() if isinstance(items, dict) else items

    def __iter__(self):
        output = list()
        for i, kv in enumerate(self.items):
            output.append((i, *kv))
        return iter(output)


def selective_slice(input, *keys, only_values: bool = False, is_tuple: bool = False) -> tdl:
    output = dict()
    for key in keys:
        output[key] = input[key]
    if only_values:
        return _is_tuple(output.values(), is_tuple)
    return output


def dict_get(dct: dict, key, default=None):
    output = dct.get(key)
    if output and output is not False:
        return default
    return output


def mislist(input: tuplist, *args, is_tuple: bool = False) -> tuplist:
    output = list()
    for arg in args:
        if arg not in input:
            output.append(arg)
    return _is_tuple(output, is_tuple)


def dict_slice(input: dict, *keys, only_values: bool = False, is_tuple: bool = False, all_required: bool = False) -> tdl:
    output = dict()
    for key in keys:
        if key in input:
            output[key] = input[key]
        elif all_required:
            raise KeyError(key)
    return _is_tuple(output.values(), is_tuple) if only_values else output



from typing import Optional, Union
from kivy.graphics import Color, Rectangle
from kivy.uix.button import Button
from kivy.uix.dropdown import DropDown
from kivy.uix.gridlayout import GridLayout
from kivy.uix.popup import Popup
from kivy.uix.spinner import Spinner as KivyComboBox
from kivy.uix.togglebutton import ToggleButton as KivyToggleButton
from kivy.uix.checkbox import CheckBox as KivyCheckBox
from kivy.uix.textinput import TextInput as KivyTextInput
from hrenpack.framework.kivy import rgba
from hrenpack.framework.kivy.mixins import WidgetMixin
from hrenpack.listwork import key_in_dict

_rgb_type = Union[tuple[int, int, int], tuple[int, int, int, int]]


class ComboBox(KivyComboBox, WidgetMixin):
    def __init__(self, parent=None, *elements, index: int = 0, command=None, **kwargs):
        elements = list(elements)
        if not key_in_dict(kwargs, 'text'):
            kwargs['text'] = elements[index]
        super(ComboBox, self).__init__(values=elements, **kwargs)
        if parent is not None:
            parent.add_widget(self)
        if command is not None:
            self.connect(command)

    def connect(self, command):
        self.bind(text=command)

    def append(self, element: str):
        self.values.append(element)

    def index(self, element: str):
        return self.values.index(element)

    def get_selected(self):
        try:
            return self.index(self.text)
        except ValueError:
            return None

    def set_selected(self, index: int):
        self.text = self.values[index]


class CustomBackgroundComboBox(ComboBox):
    """Находится в разработке, используйте на свой страх и риск"""
    class _DropDown(DropDown):
        def __init__(self, parent, background_color, **kwargs):
            super().__init__(**kwargs)
            self._background_color = background_color
            self.parent_ = parent
            with self.canvas.before:
                Color(*self.background_color)
                self.rect = Rectangle(size=self.size, pos=self.pos)

            self.bind(size=self._update_rect, pos=self._update_rect)

        @property
        def background_color(self):
            return rgba(*self._background_color)

        @background_color.setter
        def background_color(self, color: _rgb_type):
            self._background_color = color
            with self.canvas.before:
                Color(*self.background_color)

        def _update_rect(self, instance, value):
            self.rect.pos = self.pos
            self.rect.size = self.size

    def __init__(self,  parent=None, background_color: Optional[_rgb_type] = None, *elements, index: int = 0,
                 command=None, **kwargs):
        super(CustomBackgroundComboBox, self).__init__(parent, *elements, index=index, command=command, **kwargs)
        self._dropdown = self._DropDown(self, background_color)
        self._background_color = background_color if background_color is not None else super().background_color

    def _on_drop_down(self, combobox, dropdown):
        with combobox.canvas.before:
            Color(*self.background_color)
            Rectangle(size=combobox.size, pos=combobox.pos)

    @property
    def background_color(self):
        return rgba(*self._background_color)

    @background_color.setter
    def background_color(self, color: _rgb_type):
        self._background_color = color
        self._dropdown.background_color = color


class ToggleButtonGroup:
    def __init__(self, name: str, parent=None, command=None, **kwargs):
        self._name = name
        self._buttons = list()
        self._command = command
        self._auto_add_to_parent = kwargs.get('auto_add_to_parent', False)
        self.parent = parent

    def __call__(self, button):
        self._buttons.append(button)
        if self._command is not None:
            button.connect(self._command)
        if button.name is None:
            button.name = str(len(self._buttons) - 1)
        if self.parent is not None and self._auto_add_to_parent:
            self.parent.add_widget(button)
        return self.name

    def __str__(self):
        return self._name

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, new_name: str):
        self._name = new_name

    def buttons(self) -> list:
        return self._buttons

    def connect(self, command=None):
        if command is None and self._command is not None:
            command = self._command
        elif command is None:
            raise ValueError('Command is None')
        for button in self.buttons():
            button.bind(on_press=command)

    def add_to_parent(self):
        for button in self.buttons():
            self.parent.add_widget(button)

    def get_selected(self):
        for button in self.buttons():
            if button.state == 'down':
                return button.name
        else:
            return None

    def set_selected(self, name):
        for button in self.buttons():
            if button.state == 'down':
                button.state = 'normal'
                continue
            if button.name == name:
                button.state = 'down'
                break
        else:
            raise ValueError(f'No such button: {name}')


class ToggleButton(KivyToggleButton, WidgetMixin):
    _name = None

    def __init__(self, text: str = '', group: Optional[ToggleButtonGroup] = None, parent=None,
                 name: Optional[str] = None, **kwargs):
        super(ToggleButton, self).__init__(text=text, group=group(self), **kwargs)
        if parent is not None:
            parent.add_widget(self)
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, new_name: str):
        self._name = new_name

    def __str__(self):
        return "ToggleButton('{}')".format(self.name)

    def connect(self, command):
        """Не рекомендуется к использованию, лучше использовать group.connect(command)"""
        self.bind(on_press=command)


class ToggleButtonWithoutParentInInit(ToggleButton):
    def __init__(self, text: str = '', group: Optional[ToggleButtonGroup] = None, name: Optional[str] = None, **kwargs):
        super().__init__(text, group, name=name, **kwargs)


class CheckBox(KivyCheckBox, WidgetMixin):
    def __init__(self, parent=None, command=None, checked: bool = False, **kwargs):
        super(CheckBox, self).__init__(**kwargs)
        if parent is not None:
            parent.add_widget(self)
        if command is not None:
            self.connect(command)
        if checked:
            self.active = True
        self.get_selected = self.is_selected

    def connect(self, command):
        self.bind(active=command)

    def is_selected(self):
        return self.active

    def set_selected(self, active: Optional[bool] = None):
        if active is None:
            self.active = not self.active
        else:
            self.active = active


class TextInput(KivyTextInput, WidgetMixin):
    def __init__(self, parent=None, command=None, placeholder: str = '', multiline: bool = False, **kwargs):
        super().__init__(hint_text=placeholder, multiline=multiline, **kwargs)
        if parent is not None:
            parent.add_widget(self)
        if command is not None:
            self.connect(command)

    @property
    def placeholder(self) -> str:
        return self.hint_text

    @placeholder.setter
    def placeholder(self, new_placeholder: str):
        self.hint_text = new_placeholder

    def get_text(self):
        return self.text

    def set_text(self, text):
        self.text = text

    def connect(self, command):
        self.bind(text=command)



import platform, ctypes
from ctypes import wintypes
from typing import Union
from screeninfo import get_monitors


class ResolutionNotFoundError(Exception):
    pass


def get_monitor_info(monitor: int = 0):
    return get_monitors()[monitor]


def get_primary_monitor_info():
    for monitor in get_monitors():
        if monitor.is_primary:
            return monitor


def get_resolution(monitor: int = 0, letter: str = '') -> Union[tuple[int, int], str]:
    monitor_info = get_monitor_info(monitor)
    width, height = monitor_info.width, monitor_info.height
    return (width, height) if letter == '' else f'{width}{letter}{height}'


def resolution_function(**resolutions):
    """kwarg example: r1920x1080=func"""
    resolution = get_resolution(letter='x')
    rr = 'r' + resolution
    if rr in resolutions:
        resolutions[rr]()
    elif 'default' in resolutions:
        resolutions['default']()
    else:
        raise ResolutionNotFoundError('Вашего разрешения нет в списке')


def resolution_return(**resolutions):
    """kwarg example: r1920x1080=value"""
    resolution = get_resolution(letter='x')
    rr = 'r' + resolution
    if rr in resolutions:
        return resolutions[rr]
    elif 'default' in resolutions:
        return resolutions['default']
    else:
        raise ResolutionNotFoundError('Вашего разрешения нет в списке')


def get_refresh_rate():
    if platform.system() == 'Windows':
        import win32api
        import win32con
        return win32api.EnumDisplaySettings(None, win32con.ENUM_CURRENT_SETTINGS).DisplayFrequency
    else:
        raise OSError("Данная функция работает только на Windows")



from typing import Literal
from hrenpack.framework.kivy.widgets import BoxLayout, Label


class _WrapperLayout(BoxLayout):
    def __init__(self, orientation: Literal['vertical', 'horizontal'], widget, padding=0, spacing=0, **kwargs):
        super().__init__(orientation, padding=padding, spacing=spacing, **kwargs)
        self.widget = widget
        self.add_widget(self.widget)

    def get_child(self):
        return self.widget


class BoxLabel(_WrapperLayout):
    def __init__(self, orientation: Literal['vertical', 'horizontal'], padding=0, spacing=0, text: str = '', **kwargs):
        box_layout_kwargs: dict = kwargs.get('box_layout_kwargs', dict())
        label_kwargs: dict = kwargs.get('label_kwargs', dict())
        self.size_hint_y = kwargs.get('size_hint_y', None)
        super().__init__(orientation, Label(text, **label_kwargs), padding, spacing, **box_layout_kwargs)





import pygame, sys


class Keyboard:
    A = pygame.K_a
    B = pygame.K_b
    C = pygame.K_c
    D = pygame.K_d
    E = pygame.K_e
    F = pygame.K_f
    G = pygame.K_g
    H = pygame.K_h
    I = pygame.K_i
    J = pygame.K_j
    K = pygame.K_k
    L = pygame.K_l
    M = pygame.K_m
    N = pygame.K_n
    O = pygame.K_o
    P = pygame.K_p
    Q = pygame.K_q
    R = pygame.K_r
    S = pygame.K_s
    T = pygame.K_t
    U = pygame.K_u
    V = pygame.K_v
    W = pygame.K_w
    X = pygame.K_x
    Y = pygame.K_y
    Z = pygame.K_z
    SPACE = pygame.K_SPACE


class Image:
    def __init__(self, path):
        self.image = pygame.image.load(path)

    def __call__(self):
        return self.image

    def resize(self, width, height):
        self.image = pygame.transform.scale(self.image, (width, height))
        return self.image

    def resize_and_convert_alpha(self, width, height):
        self.image = self.resize(width, height).convert_alpha()
        return self.image

    @classmethod
    def quick_resize(cls, path, width, height):
        return cls(path).resize(width, height)

    @classmethod
    def quick_resize_and_convert_alpha(cls, path, width, height):
        return cls(path).resize_and_convert_alpha(width, height)


def quit():
    pygame.quit()
    sys.exit()


def quit_if_quit(event):
    if event.type == pygame.QUIT:
        quit()




from typing import Union, Optional, Literal
from pathlib import Path
from pathlike_typing import PathLike
from hrenpack.classes import range_plus


def literal_add(base, *args):
    return Union[base, Literal[*args]]


Number = Union[int, float]
SimpleList = Union[list, tuple, set]
si = Union[int, str]
IntStr = si
NullStr = Optional[str]
integer, string, boolean = int, str, bool
ColorTyping = Union[tuple[int, int, int], list[int, int, int], tuple[int, int, int, float], list[int, int, int, float]]
FivePointScale = Literal[*range_plus(5)]
TenPointScale = Literal[*range_plus(10)]
ZeroFivePointScale = literal_add(FivePointScale, 0)
ZeroTenPointScale = literal_add(TenPointScale, 0)
ThemeType = Literal['light', 'dark']



import inspect, functools
from django.core.exceptions import ImproperlyConfigured
from django.http import JsonResponse


def post_json_response(view):
    @functools.wraps(view)
    def wrapper(request, *args, **kwargs):
        if inspect.ismethod(view):
            raise ImproperlyConfigured('This view supports only FBV')
        elif request.method == 'POST':
            return view(request, *args, **kwargs)
        return JsonResponse(dict(
            success=False,
            message='Required POST method',
            code=405
        ), status=405)
    return wrapper



import mimetypes
from hrenpack.constants import MIME_TYPES, COMPOUND_EXTENSIONS


def get_mime_type(path: str):
    return mimetypes.guess_type(path)[0] or 'application/octet-stream'


def get_mime_type_extension(path: str):
    pass


def get_mime_type_filetype(path: str):
    from filetype import guess
    kind = guess(path)
    if kind is None:
        return 'application/octet-stream'
    return kind.mime


def get_mime_type_magic(path: str):
    from puremagic import from_file
    return from_file(path, True)


if __name__ == '__main__':
    print(get_mime_type('d.cpp'))



import ui
from PySide6.QtWidgets import QDialog
from PySide6.QtGui import QIcon
from hrenpack.framework.Pyside6 import exit
from hrenpack.filework import ConfigurationFile
from hrenpack.kwargswork import kwarg_return


class DialogDefault(QDialog):
    def __init__(self, config: ConfigurationFile, icon_path: str, parent=None, **kwargs):
        super().__init__(parent)
        self.ui = ui.default_ui.Ui_Default()
        self.ui.setupUi(self)
        self.setWindowIcon(QIcon(icon_path))
        self.config = config
        self.ui.label.setText(kwarg_return(kwargs, 'label_text', self.ui.label.text()))
        self.sections: dict = kwarg_return(kwargs, 'sections', {})
        self.exit = lambda: exit(self)
        self.ui.btn_yes.clicked.connect(lambda: self.default(kwarg_return(kwargs, 'delete', False)))
        self.ui.btn_no.clicked.connect(self.exit)

    def default(self, delete: bool = False):
        if self.sections:
            for section, block in self.sections.items():
                self.config.edit_section(section, block, delete)
        self.exit()



from abc import ABC, ABCMeta, abstractmethod
from typing import Literal
from kivy.uix.widget import WidgetMetaclass
from hrenpack.type_define import is_object
from hrenpack.framework.kivy.screenmanager import Screen
from hrenpack.framework.kivy.mixins import WidgetMixin
from hrenpack.framework.kivy.widgets import BoxLayout, FloatLayout

__all__ = ['MultiScreenModeFloatLayout', 'MultiScreenModeBoxLayout', 'MultiScreenModeScreen']


class MultiScreenModeMeta(ABCMeta, WidgetMetaclass):
    """Metaclass for MultiScreenMode widgets"""
    pass


class MultiScreenModeBoxLayout(ABC, BoxLayout, metaclass=MultiScreenModeMeta):
    @abstractmethod
    def __init__(self, app, screen, orientation: Literal['vertical', 'horizontal'] = 'vertical',
                 with_view: bool = False, **kwargs):
        super().__init__(orientation, **kwargs)
        self.app = app
        self.screen = screen
        if not with_view:
            self.screen.add_widget(self)


class MultiScreenModeFloatLayout(ABC, FloatLayout, metaclass=MultiScreenModeMeta):
    @abstractmethod
    def __init__(self, app, screen, with_view: bool = False, **kwargs):
        super().__init__(**kwargs)
        self.app = app
        self.screen = screen
        if not with_view:
            self.screen.add_widget(self)


class MultiScreenModeScreen(ABC, Screen, metaclass=MultiScreenModeMeta):
    @abstractmethod
    def __init__(self, app, child, orientation: Literal['vertical', 'horizontal'] = 'vertical',
                 with_view: bool = False, **kwargs):
        super().__init__(**kwargs)
        self.app = app
        if issubclass(child, MultiScreenModeBoxLayout) and not is_object(child):
            self.layout = child(app, self, orientation=orientation, with_view=with_view)


# def get_app(widget, max_path: int = 20):
#     ps = 0
#     if not isinstance(widget, WidgetMixin):
#         raise TypeError(f'widget must be of type WidgetMixin')
#     while True:
#         try:
#             if ps == 0:
#                 return widget.app
#             elif ps > max_path:
#                 raise RecursionError('Max path reached')
#             else:
#                 return eval(f"widget.{'.'.join(['parent'] * ps)}.app")
#         except AttributeError:
#             ps += 1



import encodings, chardet
from typing import Union, Optional
from charset_normalizer import from_bytes
from hrenpack.listwork import replace_fragment_from_args

ALL_ENCODINGS = set(replace_fragment_from_args('_', '-', *set(encodings.aliases.aliases.values())))


def test_file_encoding(path: str, encoding: str, message: bool = False):
    try:
        with open(path, encoding=encoding) as file:
            try:
                file.read()
            except Exception as err:
                if message:
                    return f"Ошибка {err.__class__.__name__} в кодировке", encoding
                return False
            else:
                if message:
                    return "С кодировкой", encoding, "все нормально"
                return True
    except Exception as error:
        if message:
            return f"Ошибка {error.__class__.__name__} с открытием файла с кодировкой", encoding
        return False


def select_encoding(path: str) -> str:
    for enc in ALL_ENCODINGS:
        if test_file_encoding(path, enc):
            return enc


def test_all_encodings(path: str):
    for enc in ALL_ENCODINGS:
        print(test_file_encoding(path, enc, True))


def get_encoding(input: Union[str, bytes, bytearray]) -> str:
    return from_bytes(input).best().encoding


def convert_to_utf_8(input: bytes, encoding: Optional[str] = None) -> bytes:
    if encoding is None:
        encoding = get_encoding(input)
    return input.decode(encoding).encode('utf-8')


def detect_encoding(file_content):
    # Сначала проверяем явные маркеры (BOM)
    if file_content.startswith(b'\xef\xbb\xbf'):
        return 'utf-8-sig'
    elif file_content.startswith(b'\xff\xfe'):
        return 'utf-16'

    # Используем chardet для сложных случаев
    result = chardet.detect(file_content)
    encoding = result['encoding'].lower()

    # Корректируем типичные ошибки
    encoding_map = {
        'windows-1251': 'cp1251',
        'utf_8': 'utf-8',
        'ascii': 'utf-8'  # ASCII - подмножество UTF-8
    }
    return encoding_map.get(encoding, encoding)



import re
from django.urls import get_resolver


def get_all_app_prefixes():
    """
    Получить словарь всех приложений и их префиксов
    """
    resolver = get_resolver()
    app_prefixes = {}

    for pattern in resolver.url_patterns:
        if hasattr(pattern, 'url_patterns'):
            pattern_str = str(pattern)
            match = re.search(r"include\('([\w\.]+)'\)", pattern_str)
            if match:
                app_path = match.group(1)
                app_name = app_path.split('.')[0]
                app_prefixes[app_name] = str(pattern.pattern).rstrip('/')

    return app_prefixes